from typing import Sequence

import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from ....utils.constants import (
    CM_AS_INCH,
    FIGURE_HEIGHT_CURTAIN,
    FIGURE_HEIGHT_LINE,
    FIGURE_HEIGHT_SWATH,
    FIGURE_MAP_WIDTH,
    FIGURE_WIDTH_CURTAIN,
    FIGURE_WIDTH_PROFILE,
)
from ..figure_type import FigureType


def create_fig_layout_map_main_zoom_profile(
    main_rows: Sequence[FigureType | int],
    zoom_rows: Sequence[FigureType | int] | None = None,
    profile_rows: Sequence[FigureType | int] | None = None,
    map_rows: Sequence[FigureType | int] | None = None,
    wspace: float | Sequence[float] = 3.0 * CM_AS_INCH,
    hspace: float | Sequence[float] = 3.0 * CM_AS_INCH,
    wmain: float = FIGURE_WIDTH_CURTAIN,
    hrow: float = FIGURE_HEIGHT_CURTAIN,
    hswath: float = FIGURE_HEIGHT_SWATH,
    hline: float = FIGURE_HEIGHT_LINE,
    wprofile: float = FIGURE_WIDTH_PROFILE,
    wmap: float = FIGURE_MAP_WIDTH,
    wzoom: float = FIGURE_WIDTH_CURTAIN / 3.0,
) -> tuple[Figure, Sequence[Axes], Sequence[Axes], Sequence[Axes], Sequence[Axes]]:
    """
    Creates a complex figure layout with columns for map, main, zoom, and profile panels (in that order from left to right).

    Each panel column can have a custom sequence of figure types (e.g., row heights), and the layout
    supports both uniform and per-gap horizontal/vertical spacing.

    Args:
        main_rows (Sequence[FigureType | int]): List of figure types for the rows of the main column.
        zoom_rows (Sequence[FigureType | int], optional): List of figure types for the rows in the optional zoom column.
        profile_rows (Sequence[FigureType | int], optional): List of figure types for the rows in the optional profile column.
        map_rows (Sequence[FigureType | int], optional): List of figure types for the rows in the optional map column.
        wspace (float | Sequence[float], optional): Horizontal spacing between columns. Can be a single value
            or list defining spacing before, between, and after columns.
        hspace (float | Sequence[float], optional): Vertical spacing between rows. Similar behavior as `wspace`.
        wmain (float, optional): Width of the main column. Default is `FIGURE_WIDTH_CURTAIN`.
        hrow (float, optional): Height of a standard row. Default is `FIGURE_HEIGHT_CURTAIN`.
        hswath (float, optional): Height of a `SwathFigure`-type row. Default is `FIGURE_HEIGHT_SWATH`.
        wprofile (float, optional): Width of the profile column.
        wmap (float, optional): Width of the map column.
        wzoom (float, optional): Width of the zoom column.

    Returns:
        tuple: A tuple containing:
            - Figure: The matplotlib figure object.
            - Sequence[Axes]: Axes for map panels (may be empty).
            - Sequence[Axes]: Axes for main panels.
            - Sequence[Axes]: Axes for zoom panels (may be empty).
            - Sequence[Axes]: Axes for profile panels (may be empty).

    Raises:
        ValueError: If the provided spacing sequences are of invalid length.
        TypeError: If spacing arguments are of unsupported types.
    """
    # Calculate number of columns
    is_map_col: bool = isinstance(map_rows, list) and len(map_rows) > 0
    is_main_col: bool = isinstance(main_rows, list) and len(main_rows) > 0
    is_zoom_col: bool = isinstance(zoom_rows, list) and len(zoom_rows) > 0
    is_profile_col: bool = isinstance(profile_rows, list) and len(profile_rows) > 0
    col_present: list[bool] = [is_map_col, is_main_col, is_zoom_col, is_profile_col]

    ncols: int = sum(col_present)

    # Calculate number of rows
    nrows_min: int = len(main_rows)
    if isinstance(map_rows, list):
        for ft in map_rows:
            nrows_min += 1
            if ft == FigureType.MAP_2_ROW:
                nrows_min += 1

    nrows: int = max(nrows_min, len(main_rows))

    # Calulate spaces between figures
    def _calulate_spaces(
        space: float | Sequence[float],
        n: int,
        name: str,
        name_col_row: str,
    ) -> list[float]:
        if isinstance(space, Sequence):
            space = list(space)
            if len(space) < n - 1 or len(space) > n + 1:
                raise ValueError(
                    f"{name} was given as a list (size={len(space)}) and thus needs to have a size between number of {name_col_row} ({n}) -1 (i.e. only spaces between {name_col_row}) and +1 (i.e. spaces before, between and after {name_col_row})."
                )
            elif len(space) == n - 1:
                space = [0.0] + space + [0.0]
            elif len(space) == n:
                space = space + [0.0]
        elif isinstance(space, float):
            space = [0.0] + [space] * (n - 1) + [0.0]
        else:
            raise TypeError(
                f"{name} has wrong type '{type(space).__name__}'. expected types: '{float.__name__}' or ''{list.__name__}'[{float.__name__}]'"
            )
        return space

    wspace = _calulate_spaces(wspace, ncols, "wspace", "columns")
    hspace = _calulate_spaces(hspace, nrows, "hspace", "rows")

    # Calculate size ratios of figures
    def _get_ratios(
        ratios_figs: list[float],
        space: list[float],
    ) -> list[float]:
        assert len(space) == len(ratios_figs) + 1

        ratios: list[float] = []
        for i, r in enumerate(ratios_figs):
            ratios.append(space[i])
            ratios.append(r)
        ratios.append(space[-1])

        return ratios

    wratios_figs: list[float] = np.array([wmap, wmain, wzoom, wprofile])[
        col_present
    ].tolist()
    hratios_figs: list[float] = []
    for fig_type in main_rows:
        if fig_type == FigureType.SWATH:
            hratios_figs.append(hswath)
        elif fig_type == FigureType.LINE:
            hratios_figs.append(hline)
        else:
            hratios_figs.append(hrow)
    if len(main_rows) < nrows_min:
        for i in range(nrows_min - len(main_rows)):
            hratios_figs.append(hrow)

    wratios = _get_ratios(wratios_figs, wspace)
    hratios = _get_ratios(hratios_figs, hspace)

    # Create the figure
    wfig = sum(wratios)
    hfig = sum(hratios)
    figsize = (wfig, hfig)

    fig = plt.figure(figsize=figsize)

    # Create the grid layout
    gs = gridspec.GridSpec(
        nrows=len(hratios),
        ncols=len(wratios),
        width_ratios=wratios,
        height_ratios=hratios,
        figure=fig,
        wspace=0,
        hspace=0,
        bottom=0.0,
        top=1.0,
        right=1.0,
        left=0.0,
    )

    # Create the plots
    # Create maps
    current_col: int = 1
    current_row: int = 1
    axs_map: list[Axes] = []
    axs_main: list[Axes] = []
    axs_zoom: list[Axes] = []
    axs_profile: list[Axes] = []
    ax: Axes | None
    if isinstance(map_rows, list):
        for fig_type in map_rows:
            if fig_type == FigureType.MAP_2_ROW:
                ax = fig.add_subplot(gs[current_row : current_row + 3, current_col])
                current_row += 2
            elif fig_type == FigureType.NONE:
                ax = None
            else:
                ax = fig.add_subplot(gs[current_row, current_col])
            if isinstance(ax, Axes):
                axs_map.append(ax)
            current_row += 2
        current_col += 2
        current_row = 1

    if isinstance(main_rows, list):
        for fig_type in main_rows:
            if fig_type == FigureType.NONE:
                ax = None
            else:
                ax = fig.add_subplot(gs[current_row, current_col])
            if isinstance(ax, Axes):
                axs_main.append(ax)
            current_row += 2
        current_col += 2
        current_row = 1

    if isinstance(zoom_rows, list):
        for fig_type in zoom_rows:
            if fig_type == FigureType.NONE:
                ax = None
            else:
                ax = fig.add_subplot(gs[current_row, current_col])
            if isinstance(ax, Axes):
                axs_zoom.append(ax)
            current_row += 2
        current_col += 2
        current_row = 1

    if isinstance(profile_rows, list):
        for fig_type in profile_rows:
            if fig_type == FigureType.NONE:
                ax = None
            else:
                ax = fig.add_subplot(gs[current_row, current_col])
            if isinstance(ax, Axes):
                axs_profile.append(ax)
            current_row += 2
        current_col += 2
        current_row = 1

    return fig, axs_map, axs_main, axs_zoom, axs_profile



from dataclasses import dataclass
from enum import StrEnum
from typing import Any


class AlongTrackAxisData(StrEnum):
    TIME = "time"
    TIME_UTC = "utc"
    TIME_LST = "lst"
    GEO = "geo"
    LAT = "lat"
    LON = "lon"
    DISTANCE = "distance"
    COUNT = "count"
    NONE = "none"

    @staticmethod
    def from_str(label: str) -> "AlongTrackAxisData":
        label = label.lower()
        # Convert plural to singular
        if len(label) > 0 and label[-1] == "s":
            label = label[0:-1]
        if label in (
            "latlon",
            "lonlat",
            "coord",
            "coords",
            "coordinates",
            "geo",
            "geoloc",
            "geolocation",
            "loc",
            "location",
        ):
            return AlongTrackAxisData.GEO
        elif label in ("lat", "latitude"):
            return AlongTrackAxisData.LAT
        elif label in ("lon", "longitude"):
            return AlongTrackAxisData.LON
        elif label in ("distance", "dist", "grounddist", "grounddistance"):
            return AlongTrackAxisData.DISTANCE
        elif label in ("time", "utclst", "lstutc", "timeutclst", "timelstutc"):
            return AlongTrackAxisData.TIME
        elif label in ("utc", "timeutc", "utctime"):
            return AlongTrackAxisData.TIME_UTC
        elif label in ("lst", "timelst", "lsttime"):
            return AlongTrackAxisData.TIME_LST
        elif label in ("count", "sample", "number", "num", "index", "idx"):
            return AlongTrackAxisData.COUNT
        elif label in ("none", "no", "not", "empty", "nothing", "-", ""):
            return AlongTrackAxisData.NONE
        else:
            try:
                return AlongTrackAxisData(label)
            except:
                raise ValueError(f"Can not find any match for label '{label}'")


@dataclass
class AlongTrackAxisStyle:
    data: AlongTrackAxisData
    units: bool | None = None
    title: bool | None = None
    labels: bool | None = None

    def __post_init__(self):
        self.data = AlongTrackAxisData.from_str(self.data)

    @classmethod
    def from_input(cls, input: "AlongTrackAxisStyle | str") -> "AlongTrackAxisStyle":
        if isinstance(input, cls):
            return input
        if isinstance(input, str):
            return cls(input)  # type: ignore
        raise TypeError(
            f"invalid type '{type(input).__name__}', expecting '{cls.__name__}' or 'str'"
        )



import numpy as np
from xarray import Dataset

from ...constants import ALONG_TRACK_DIM, EC_LATITUDE_FRAME_BOUNDS, TRACK_LAT_VAR
from .header_group import read_header_data


def get_frame_id(ds: Dataset) -> str:
    if "frameID" in ds:
        return str(ds.frameID.values)
    return str(read_header_data(ds).frameID.values.astype(str))


def get_frame_along_track(
    ds: Dataset, along_track_dim: str = ALONG_TRACK_DIM, lat_var: str = TRACK_LAT_VAR
) -> tuple[int, int]:
    frame_id = get_frame_id(ds)
    lat_framestart, lat_framestop = EC_LATITUDE_FRAME_BOUNDS[frame_id]

    i_halfway = len(ds[along_track_dim]) // 2
    i_framestart = np.argmin(np.abs(ds[lat_var].values[:i_halfway] - lat_framestart))
    i_framestop = i_halfway + np.argmin(
        np.abs(ds[lat_var].values[i_halfway:] - lat_framestop)
    )

    return int(i_framestart), int(i_framestop)


def trim_to_latitude_frame_bounds(
    ds: Dataset, along_track_dim: str = ALONG_TRACK_DIM, lat_var: str = TRACK_LAT_VAR
) -> Dataset:
    """
    Trims the dataset to the region within the latitude frame bounds.

    Args:
        ds (xarray.Dataset): Input dataset to be trimmed.
        along_track_dim (str, optional): Dimension along which to trim. Defaults to ALONG_TRACK_DIM.
        lat_var (str, optional): Name of the latitude variable. Defaults to TRACK_LAT_VAR.

    Returns:
        xarray.Dataset: Trimmed dataset.
    """
    return ds.isel(
        {
            along_track_dim: slice(
                *get_frame_along_track(
                    ds, along_track_dim=along_track_dim, lat_var=lat_var
                )
            )
        }
    )



import numpy as np
import xarray as xr
from numpy.typing import NDArray

from ..constants import TRACK_LAT_VAR, TRACK_LON_VAR


def get_coords(
    ds: xr.Dataset,
    *,
    lat_var: str = TRACK_LAT_VAR,
    lon_var: str = TRACK_LON_VAR,
    flatten: bool = False,
) -> NDArray:
    """Takes a xarray Dataset and returns the lat/lon coordinates as a numpy array."""
    lat = ds[lat_var]
    lon = ds[lon_var]
    coords = np.stack((lat, lon)).transpose()

    if len(coords.shape) > 2 and flatten:
        coords = coords.reshape(-1, 2)
    return coords



import functools
import os
import threading
import time
from logging import Logger
from typing import Any, Callable


def log_message(msg: str, logger: Logger | None, level: str = "info") -> None:
    """Logs a message either to a logger or to stdout.

    Args:
        msg (str): The message to log.
        logger (Logger | None): Logger instance to use. If None, prints to stdout.
        level (str): Logging level as a string ("info", "warning", "error", etc.).
    """
    if logger:
        log_func = getattr(logger, level, logger.info)
        log_func(msg)
    else:
        print(msg)


def is_heartbeat_stale(heartbeat_path: str | None, max_age_seconds: int) -> bool:
    """Checks whether the heartbeat file is stale, which indicates that the process is no longer running.

    Args:
        heartbeat_path (str | None): Path to the heartbeat file.
        max_age_seconds (int): Maximum allowed age in seconds before considering it stale.

    Returns:
        True if the heartbeat is missing or stale, False otherwise.
    """
    if not heartbeat_path or not os.path.exists(heartbeat_path):
        return True  # No heartbeat = assume stale
    last_update = os.path.getmtime(heartbeat_path)
    age = time.time() - last_update
    return age > max_age_seconds


def heartbeat_writer(
    heartbeat_path: str,
    heartbeat_interval_seconds: int,
    stop_event: threading.Event,
    logger: Logger | None,
) -> None:
    """Periodically updates the heartbeat file until stopped.

    Args:
        heartbeat_path (str): Path to the heartbeat file.
        heartbeat_interval_seconds (int): Interval in seconds between heartbeat updates.
        stop_event (threading.Event): Event used to signal stopping the heartbeat writer.
        logger (Logger | None): Logger instance to use. If None, prints to stdout.
    """
    update_interval_seconds = 1  # seconds
    elapsed = 0

    while not stop_event.is_set():
        if elapsed >= heartbeat_interval_seconds:
            try:
                with open(heartbeat_path, "w") as f:
                    f.write(str(time.time()))
            except Exception as e:
                log_message(f"Failed to update heartbeat: {e}", logger, "error")
            elapsed = 0

        time.sleep(update_interval_seconds)
        elapsed += update_interval_seconds


def with_lockfile(
    lockfile_path: str,
    logger: Logger | None = None,
    heartbeat_path: str | None = None,
    heartbeat_interval_seconds: int = 60,
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """Decorator that ensures only one instance of a script runs at a time,
    using a lock file and optional heartbeat file for stale lock detection.

    Args:
        lockfile_path (str): Path to the lock file.
        logger (Logger | None): Logger instance to use for logging. If None, prints to stdout.
        heartbeat_path (str | None): Optional path to a heartbeat file that indicates script liveness.
        heartbeat_interval_seconds (int): Interval in seconds for heartbeat updates and staleness detection.

    Returns:
        A decorator that wraps the target function with lockfile and heartbeat management.
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            # Handle existing lock
            if os.path.exists(lockfile_path):
                max_heartbeat_interval_seconds = heartbeat_interval_seconds * 10
                if is_heartbeat_stale(heartbeat_path, max_heartbeat_interval_seconds):
                    log_message(
                        f"Stale lock detected (heartbeat older than {max_heartbeat_interval_seconds} seconds). Removing lock.",
                        logger,
                        "warning",
                    )
                    try:
                        os.remove(lockfile_path)
                    except Exception as e:
                        log_message(
                            f"Failed to remove stale lock: {e}", logger, "error"
                        )
                        return
                else:
                    log_message(
                        f"Script is already running (lock: {lockfile_path}, heartbeat is fresh). Exiting.",
                        logger,
                        "warning",
                    )
                    return

            # Create new lock
            try:
                open(lockfile_path, "w").close()
                log_message("Lock activated.", logger)
            except Exception as e:
                log_message(f"Failed to create lock file: {e}", logger, "error")
                return

            # Setup heartbeat
            heartbeat_thread: threading.Thread | None = None
            heartbeat_stop_event: threading.Event = threading.Event()

            if heartbeat_path:
                heartbeat_thread = threading.Thread(
                    target=heartbeat_writer,
                    args=(
                        heartbeat_path,
                        heartbeat_interval_seconds,
                        heartbeat_stop_event,
                        logger,
                    ),
                    daemon=True,
                )
                heartbeat_thread.start()

            try:
                return func(*args, **kwargs)
            finally:
                # Cleanup heartbeat
                if heartbeat_path:
                    heartbeat_stop_event.set()
                    if heartbeat_thread:
                        heartbeat_thread.join()
                    try:
                        os.remove(heartbeat_path)
                        log_message("Heartbeat deactivated.", logger)
                    except FileNotFoundError:
                        pass
                    except Exception as e:
                        log_message(f"Failed to remove heartbeat: {e}", logger, "error")

                # Cleanup lock
                try:
                    os.remove(lockfile_path)
                    log_message("Lock deactivated.", logger)
                except FileNotFoundError:
                    pass
                except Exception as e:
                    log_message(f"Failed to remove lock file: {e}", logger, "error")

        return wrapper

    return decorator



import os
import warnings
from dataclasses import dataclass, field
from enum import StrEnum
from pathlib import Path
from typing import Any, Final, Literal

from .. import __title__

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib  # type: ignore

import tomli_w

DEAULT_CONFIG_FILENAME: Final[str] = "default_config.toml"
DEFAULT_CONFIG_TEXT: Final[
    str
] = '''[local]
# Set a path to your root EarthCARE data directory,
# where local EarthCARE product files will be searched and downloaded to.
data_directory = ""

# Set a path to your root image directory,
# where saved plots will be put.
image_directory = ""

[download]
# You have 2 options to set your data access rights:
# 1. (recommended) Choose one: "commissioning", "calval" or "open", e.g.:
#         collections = "calval"
# 2. List individual collections, e.g.:
#         collections = [
#             "EarthCAREL1InstChecked",
#             "EarthCAREL2InstChecked",
#             ...
#         ]
collections = "open"

# Set your data dissemination service that will be used for remote data search and download.
# Choose one: "oads" or "maap"
platform = "oads"

# If you've choosen "maap", generate a data access token on EarthCARE MAAP and put it here:
# (see <https://portal.maap.eo.esa.int/earthcare/>)
maap_token = ""

# If you've choosen "oads", give your OADS credencials here:
# (see <https://ec-pdgs-dissemination1.eo.esa.int> and <https://ec-pdgs-dissemination2.eo.esa.int>)
oads_username = "my_username"
oads_password = """my_password"""
'''
DEFAULT_CONFIG_SETUP_INSTRUCTIONS: Final[str] = (
    "\n\tPlease create a custom configuration file (TOML).\n"
    "\tTo do this you can follow these steps:\n\n"
    "\t1. Generate a template configuration file by running in your Python code:\n"
    "\t       >>> import earthcarekit as eck\n"
    "\t       >>> eck.create_example_config(path_to_file_or_dir)\n\n"
    "\t2. Edit the fields the generated file using a text editor.\n\n"
    "\t3. Finally to run in your Python code:\n"
    "\t       >>> eck.set_config(path_to_file)\n\n"
    "\tThis will generate a file in your users home directory (see <~/.config/default_config.toml)>\n"
    f"\twhich will be used as the default configuration of '{__title__}'.\n"
)


class DisseminationCollection(StrEnum):
    """Enum for OADS data collection names."""

    EarthCAREL0L1Products = "EarthCAREL0L1Products"
    EarthCAREL1InstChecked = "EarthCAREL1InstChecked"
    EarthCAREL1Validated = "EarthCAREL1Validated"
    EarthCAREL2Products = "EarthCAREL2Products"
    EarthCAREL2InstChecked = "EarthCAREL2InstChecked"
    EarthCAREL2Validated = "EarthCAREL2Validated"
    JAXAL2Products = "JAXAL2Products"
    JAXAL2InstChecked = "JAXAL2InstChecked"
    JAXAL2Validated = "JAXAL2Validated"
    EarthCAREAuxiliary = "EarthCAREAuxiliary"
    EarthCAREXMETL1DProducts10 = "EarthCAREXMETL1DProducts10"
    EarthCAREOrbitData = "EarthCAREOrbitData"

    def to_maap(self) -> str:
        return f"{self.value}_MAAP"


@dataclass
class ECKConfig:
    """Class storing earthcarekit configurations."""

    filepath: str = field(default_factory=str)
    path_to_data: str = field(default_factory=str)
    path_to_images: str = field(default_factory=str)
    oads_username: str = field(default_factory=str)
    oads_password: str = field(default_factory=str)
    collections: list[DisseminationCollection] = field(
        default_factory=lambda: [
            DisseminationCollection.EarthCAREL1Validated,
            DisseminationCollection.EarthCAREL2Validated,
            DisseminationCollection.JAXAL2Validated,
            DisseminationCollection.EarthCAREOrbitData,
        ]
    )
    maap_token: str = field(default_factory=str)
    download_backend: str = "oads"
    user_type: str = "none"

    def __repr__(self):
        data = [
            f"filepath='{self.filepath}'",
            f"path_to_data='{self.path_to_data}'",
            f"path_to_images='{self.path_to_images}'",
            f"oads_username='{self.oads_username}'",
            f"oads_password='***'",
            f"collections='{self.collections}'",
            f"maap_token='{self.maap_token}'",
        ]
        return f"{ECKConfig.__name__}({', '.join(data)})"


def get_default_config_filepath() -> str:
    user_dir = os.path.expanduser("~")
    config_dir = os.path.join(user_dir, ".config", "earthcarekit")
    config_filepath = os.path.join(config_dir, DEAULT_CONFIG_FILENAME)
    return config_filepath


def ensure_filepath(filepath: str) -> None:
    path = Path(filepath)
    path.parent.mkdir(parents=True, exist_ok=True)
    path.touch(exist_ok=True)


def get_collections_from_user_type_str(
    user_type_str: Literal["commissioning", "calval", "open", "none"],
) -> list[str]:
    colls_comm = [
        "EarthCAREL0L1Products",
        "EarthCAREL2Products",
        "JAXAL2Products",
    ]

    colls_calval = [
        "EarthCAREAuxiliary",
        "EarthCAREL1InstChecked",
        "EarthCAREL2InstChecked",
        "JAXAL2InstChecked",
    ]

    colls_public = [
        "EarthCAREL1Validated",
        "EarthCAREL2Validated",
        "JAXAL2Validated",
        "EarthCAREXMETL1DProducts10",
        "EarthCAREOrbitData",
    ]

    if user_type_str == "commissioning":
        return colls_comm + colls_calval + colls_public
    elif user_type_str == "calval":
        return colls_calval + colls_public
    elif user_type_str == "open":
        return colls_public
    elif user_type_str == "none":
        return []
    else:
        raise ValueError(
            f'invalid user_type_str "{user_type_str}". expected EarthCARE user types are: "commissioning", "calval", "open" or "none"'
        )


def read_config(config_filepath: str | None = None) -> ECKConfig:
    """Reads and return earthcare-kit configurations."""
    if isinstance(config_filepath, str):
        config_filepath = os.path.abspath(config_filepath)
    elif config_filepath is None:
        config_filepath = get_default_config_filepath()
    else:
        raise TypeError(
            f"got invalid type for `path_to_config` ('{type(config_filepath).__name__}'), expected type 'str'"
        )

    if os.path.exists(config_filepath):
        with open(config_filepath, "rb") as f:
            config = tomllib.load(f)
            try:
                if "Local_file_system" in config:
                    data_dirpath = config["Local_file_system"]["data_directory"]
                    image_dirpath = config["Local_file_system"]["image_directory"]
                else:
                    data_dirpath = config["local"]["data_directory"]
                    image_dirpath = config["local"]["image_directory"]

                download_backend: str
                user_type: Literal["commissioning", "calval", "open", "none"] = "none"
                collections: str | list[str] | None
                if "OADS_credentials" in config:
                    oads_username = config.get("OADS_credentials", dict).get(
                        "username", ""
                    )
                    oads_password = config.get("OADS_credentials", dict).get(
                        "password", ""
                    )
                    collections = config.get("OADS_credentials", dict).get(
                        "collections", None
                    )
                    download_backend = config.get("OADS_credentials", dict).get(
                        "platform", "oads"
                    )
                    maap_token = config.get("OADS_credentials", dict).get(
                        "maap_token", ""
                    )
                else:
                    oads_username = config.get("download", dict).get(
                        "oads_username", ""
                    )
                    oads_password = config.get("download", dict).get(
                        "oads_password", ""
                    )
                    collections = config.get("download", dict).get("collections", None)
                    download_backend = config.get("download", dict).get(
                        "platform", "oads"
                    )
                    maap_token = config.get("download", dict).get("maap_token", "")

                if isinstance(collections, str):
                    if collections.lower() == "commissioning":
                        user_type = "commissioning"
                        collections = get_collections_from_user_type_str(user_type)
                    elif collections.lower() == "calval":
                        user_type = "calval"
                        collections = get_collections_from_user_type_str(user_type)
                    elif collections.lower() == "open":
                        user_type = "open"
                        collections = get_collections_from_user_type_str(user_type)

                _collections: list[DisseminationCollection] = []
                if isinstance(collections, list):
                    _collections = [DisseminationCollection(c) for c in collections]

                eckit_config = ECKConfig(
                    filepath=config_filepath,
                    path_to_data=data_dirpath,
                    path_to_images=image_dirpath,
                    oads_username=oads_username,
                    oads_password=oads_password,
                    collections=_collections,
                    maap_token=maap_token,
                    download_backend=download_backend.lower(),
                    user_type=user_type,
                )
                return eckit_config
            except AttributeError as e:
                raise AttributeError(f"Invalid config file is missing variable: {e}")

    raise FileNotFoundError(
        f"Missing config.toml file ({config_filepath})\n"
        f"{DEFAULT_CONFIG_SETUP_INSTRUCTIONS}"
    )


def set_config(c: str | ECKConfig, verbose: bool = True) -> None:
    _config: ECKConfig
    if isinstance(c, str):
        _config = read_config(c)
    elif isinstance(c, ECKConfig):
        _config = c
    else:
        raise TypeError(
            f"Invalid config! Either give a path to a eckit config TOML file or pass a instance of the class '{ECKConfig.__name__}'"
        )

    config = {
        "local": {
            "data_directory": _config.path_to_data,
            "image_directory": _config.path_to_images,
        },
        "download": {
            "collections": [str(oads_c) for oads_c in _config.collections],
            "platform": _config.download_backend,
            "maap_token": _config.maap_token,
            "oads_username": _config.oads_username,
            "oads_password": _config.oads_password,
        },
    }

    config_filepath = get_default_config_filepath()
    ensure_filepath(config_filepath)

    with open(config_filepath, "wb") as f:
        tomli_w.dump(config, f)

    if verbose:
        print(f"Default configuration file set at <{config_filepath}>")


def create_example_config(target_dirpath: str = ".", verbose: bool = True) -> None:
    filename: str
    dirpath: str = os.path.abspath(target_dirpath)
    if not os.path.isdir(dirpath):
        filename = os.path.basename(dirpath)
        dirpath = os.path.dirname(dirpath)
    else:
        filename = "example_config.toml"

    filepath: str = os.path.join(dirpath, filename)

    config_str = DEFAULT_CONFIG_TEXT

    with open(filepath, "w", encoding="utf-8") as f:
        f.write(config_str)

    if verbose:
        print(f"Example configuration file created at <{filepath}>")


def _warn_user_if_not_default_config_exists() -> None:
    if not os.path.exists(get_default_config_filepath()):
        msg: str = (
            f"Configuration of '{__title__}' is incomplete.\n"
            f"{DEFAULT_CONFIG_SETUP_INSTRUCTIONS}"
        )
        warnings.warn(message=msg)



import argparse
import datetime
import sys
from argparse import RawTextHelpFormatter
from typing import Any, Sequence, Type, TypeAlias

import numpy as np
import pandas as pd

from .. import __title__, __version__
from ..utils._cli import console_exclusive_info, create_logger, log_textbox
from ._constants import PROGRAM_DESCRIPTION, PROGRAM_NAME, PROGRAM_SETUP_INSTRUCTIONS
from ._create_search_requests import create_search_request_list
from ._eo_product import EOProduct, _DownloadResult
from ._eo_search_request import EOSearchRequest
from ._parse import (
    parse_path_to_config,
    parse_path_to_data,
    parse_search_inputs,
    parse_selected_index,
)
from ._run_downloads import run_downloads
from ._run_search_requests import run_search_requets
from ._types import Entrypoint, _SearchInputs

RadiusKMFloat: TypeAlias = float
LatFloat: TypeAlias = float
LonFloat: TypeAlias = float

LatSFloat: TypeAlias = float
LonWFloat: TypeAlias = float
LatNFloat: TypeAlias = float
LonEFloat: TypeAlias = float


def ecdownload(
    file_type: str | list[str],
    product_version: str | None = None,
    orbit_number: int | list[int] | None = None,
    start_orbit_number: int | None = None,
    end_orbit_number: int | None = None,
    frame_id: str | list[str] | None = None,
    orbit_and_frame: str | list[str] | None = None,
    start_orbit_and_frame: str | None = None,
    end_orbit_and_frame: str | None = None,
    timestamps: str | list[str] | None = None,
    start_time: str | None = None,
    end_time: str | None = None,
    radius_search: tuple[RadiusKMFloat, LatFloat, LonFloat] | list | None = None,
    bounding_box: (
        tuple[LatSFloat, LonWFloat, LatNFloat, LonEFloat] | list | None
    ) = None,
    path_to_config: str | None = None,
    path_to_data: str | None = None,
    is_log: bool = False,
    is_debug: bool = False,
    is_download: bool = True,
    is_overwrite: bool = False,
    is_unzip: bool = True,
    is_delete: bool = True,
    is_create_subdirs: bool = True,
    is_export_results: bool = False,
    idx_selected_input: int | None = None,
) -> None:
    time_start_script: pd.Timestamp = pd.Timestamp(
        datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    )

    def _to_list(input: Any, _type: Type) -> list | None:
        if isinstance(input, _type):
            return [input]
        elif isinstance(input, list):
            return input
        else:
            return None

    _file_type: list[str] | None = _to_list(file_type, str)
    assert isinstance(_file_type, list)
    file_type = _file_type

    orbit_number = _to_list(orbit_number, int)
    frame_id = _to_list(frame_id, str)
    orbit_and_frame = _to_list(orbit_and_frame, str)
    timestamps = _to_list(timestamps, str)

    if isinstance(radius_search, tuple):
        radius_search = list(radius_search)

    if isinstance(bounding_box, tuple):
        bounding_box = list(bounding_box)

    idx_selected: int | None = parse_selected_index(idx_selected_input)

    logger = create_logger(
        logger_name=PROGRAM_NAME,
        log_to_file=is_log,
        debug=is_debug,
    )

    log_textbox(
        f"EarthCARE Download Tool\n{__title__} {__version__}",
        logger=logger,
        is_mayor=True,
    )

    if logger:
        logger.info(f"# Settings")
        logger.info(f"# - {is_download=}")
        logger.info(f"# - {is_overwrite=}")
        logger.info(f"# - {is_unzip=}")
        logger.info(f"# - {is_delete=}")
        logger.info(f"# - {is_create_subdirs=}")
        logger.info(f"# - {is_log=}")
        logger.info(f"# - {is_debug=}")
        logger.info(f"# - {is_export_results=}")
        logger.info(f"# - {idx_selected_input=}")

    config = parse_path_to_config(path_to_config, logger=logger)
    path_to_data = parse_path_to_data(path_to_data, logger=logger)
    if isinstance(path_to_data, str):
        config.path_to_data = path_to_data

    if logger:
        logger.info(f"# - config_filepath=<{config.filepath}>")
        logger.info(f"# - data_dirpath=<{config.path_to_data}>")

    search_inputs: _SearchInputs = parse_search_inputs(
        product_type=file_type,
        product_version=product_version,
        orbit_number=orbit_number,
        start_orbit_number=start_orbit_number,
        end_orbit_number=end_orbit_number,
        frame_id=frame_id,
        orbit_and_frame=orbit_and_frame,
        start_orbit_and_frame=start_orbit_and_frame,
        end_orbit_and_frame=end_orbit_and_frame,
        timestamps=timestamps,
        start_time=start_time,
        end_time=end_time,
        radius_search=radius_search,
        bounding_box=bounding_box,
        logger=logger,
    )
    if config.download_backend.lower() == "maap":
        entrypoint = Entrypoint.MAAP
    else:
        entrypoint = Entrypoint.OADS

    planned_requests: list[EOSearchRequest] = create_search_request_list(
        entrypoint=entrypoint,
        search_inputs=search_inputs,
        input_user_type=None,
        candidate_coll_names_user=[c.value for c in config.collections],
        logger=logger,
    )

    found_products: list[EOProduct] = run_search_requets(
        log_heading_msg=f"STEP 1/2 - Search products",
        search_requests=planned_requests,
        is_debug=is_debug,
        is_found_files_list_to_txt=is_export_results,
        selected_index=idx_selected,
        selected_index_input=idx_selected_input,
        logger=logger,
    )

    donwload_results: list[_DownloadResult] = run_downloads(
        log_heading_msg=f"STEP 2/2 - Download products",
        products=found_products,
        config=config,
        entrypoint=entrypoint,
        is_download=is_download,
        is_overwrite=is_overwrite,
        is_unzip=is_unzip,
        is_delete=is_delete,
        is_create_subdirs=is_create_subdirs,
        logger=logger,
    )

    if logger:
        num_downloads: int = 0
        num_unzips: int = 0
        num_errors: int = 0
        size_msg: str = "<missing size_msg>"
        avg_speed_mbs: float = 0.0
        if len(donwload_results) > 0:
            num_errors = sum([not r.success for r in donwload_results])
            num_downloads = sum([r.downloaded for r in donwload_results])
            num_unzips = sum([r.unzipped for r in donwload_results])
            total_size_mb = sum([r.size_mb for r in donwload_results])
            size_msg = f"{total_size_mb:.2f} MB"
            if total_size_mb >= 1024:
                size_msg = f"{total_size_mb / 1024:.2f} GB"
            avg_speed_mbs = float(np.mean([r.speed_mbs for r in donwload_results]))

        time_end_script: pd.Timestamp = pd.Timestamp(
            datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        )
        execution_time: pd.Timedelta = time_end_script - time_start_script
        execution_time_str = str(execution_time).split()[-1]

        console_exclusive_info()
        _msg = [
            f"EXECUTION SUMMARY",
            "---",
            f"Time taken          {execution_time_str}",
            f"API search requests {len(planned_requests)}",
            f"Remote files found  {len(found_products)}",
            f"Files downloaded    {num_downloads} ({size_msg} at ~{avg_speed_mbs:.2f} MB/s)",
            f"Files unzipped      {num_unzips}",
            f"Errors occured      {num_errors}",
        ]
        log_textbox("\n".join(_msg), logger=logger, show_time=True)


def cli_tool_ecdownload() -> None:
    parser = argparse.ArgumentParser(
        prog=PROGRAM_NAME,
        description=f"{PROGRAM_DESCRIPTION}\n\n{PROGRAM_SETUP_INSTRUCTIONS}",
        formatter_class=RawTextHelpFormatter,
    )
    parser.add_argument(
        "product_type",
        type=str,
        nargs="*",
        help="A list of EarthCARE product names (e.g. ANOM or ATL-NOM-1B, etc.).\nYou can also specify the product version by adding a colon and the two-letter\nprocessor baseline after the name (e.g. ANOM:AD).",
    )
    parser.add_argument(
        "-d",
        "--path_to_data",
        type=str,
        default=None,
        help="The local root directory where products will be downloaded to",
    )
    parser.add_argument(
        "-o",
        "--orbit_number",
        type=int,
        nargs="*",
        default=None,
        help="A list of EarthCARE orbit numbers (e.g. 981)",
    )
    parser.add_argument(
        "-so",
        "--start_orbit_number",
        type=int,
        default=None,
        help="Start of orbit number range (e.g. 981). Can only be used in combination with option -eo.",
    )
    parser.add_argument(
        "-eo",
        "--end_orbit_number",
        type=int,
        default=None,
        help="End of orbit number range (e.g. 986). Can only be used in combination with option -so.",
    )
    parser.add_argument(
        "-f",
        "--frame_id",
        type=str,
        nargs="*",
        default=None,
        help="A EarthCARE frame ID (i.e. single letters from A to H)",
    )
    parser.add_argument(
        "-oaf",
        "--orbit_and_frame",
        type=str,
        nargs="*",
        default=None,
        help="A string describing the EarthCARE orbit number and frame (e.g. 00981E)",
    )
    parser.add_argument(
        "-soaf",
        "--start_orbit_and_frame",
        type=str,
        default=None,
        help="Start orbit number and frame range (e.g. 00981E). Can only be used in combination with option -eoaf. Can not be used with separate orbit and frame options -o, -so, eo and -f.",
    )
    parser.add_argument(
        "-eoaf",
        "--end_orbit_and_frame",
        type=str,
        default=None,
        help="End orbit number and frame range (e.g. 00982B). Can only be used in combination with option -soaf. Can not be used with separate orbit and frame options -o, -so, eo and -f.",
    )
    parser.add_argument(
        "-t",
        "--time",
        type=str,
        nargs="*",
        default=None,
        help='Search for data containing a specific timestamp (e.g. "2024-07-31 13:45" or 20240731T134500Z)',
    )
    parser.add_argument(
        "-st",
        "--start_time",
        type=str,
        default=None,
        help='Start of sensing time (e.g. "2024-07-31 13:45" or 20240731T134500Z)',
    )
    parser.add_argument(
        "-et",
        "--end_time",
        type=str,
        default=None,
        help='End of sensing time (e.g. "2024-07-31 13:45" or 20240731T134500Z)',
    )
    parser.add_argument(
        "-r",
        "--radius_search",
        type=float,
        nargs=3,
        default=None,
        help="Perform search around a radius around a point (e.g. 25000 51.35 12.43, i.e. <radius[m]> <latitude> <longitude>)",
    )
    parser.add_argument(
        "-pv",
        "--product_version",
        type=str,
        default="latest",
        help='Product version, i.e. the two-letter identifier of the processor baseline (e.g. AC). Defalut ist "latest"',
    )
    parser.add_argument(
        "-bbox",
        "--bounding_box",
        type=float,
        nargs=4,
        default=None,
        help="Perform search inside a bounding box (e.g. 14.9 37.7 14.99 37.78, i.e. <latS> <lonW> <latN> <lonE>)",
    )
    parser.add_argument(
        "--overwrite",
        action="store_true",
        help="Overwrite local data (otherwise existing local data will not be downloaded again)",
    )
    parser.add_argument(
        "--no_download", action="store_false", help="Do not download any data"
    )
    parser.add_argument(
        "--no_unzip", action="store_false", help="Do not unzip any data"
    )
    parser.add_argument(
        "--no_delete",
        action="store_false",
        help="Do not delete zip files after unzipping them",
    )
    parser.add_argument(
        "--no_subdirs",
        action="store_false",
        help="Do not create subdirs like: data_directory/data_level/product_type/year/month/day",
    )
    parser.add_argument(
        "-c",
        "--path_to_config",
        type=str,
        default=None,
        help="The path to an OADS credential TOML file (note: if not provided, a file named 'config.toml' is required in the script's folder)",
    )
    parser.add_argument(
        "--debug", action="store_true", help="Shows debug messages in console."
    )
    parser.add_argument(
        "--no_log", action="store_false", help="Prevents generation of log files."
    )
    parser.add_argument(
        "-i",
        "--select_file_at_index",
        type=int,
        default=None,
        help="Select only one product from the found products list by index for download. You may provide a negative index to start from the last entry (e.g. -1 downloads the last file listed).",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="store_true",
        help="Shows the script's version and exit",
    )
    parser.add_argument(
        "--export_results",
        action="store_true",
        help="Writes names of found files to a txt file called 'results.txt'",
    )
    args = parser.parse_args()

    if args.version:
        print(f"earthcarekit {__version__}")
        sys.exit(0)

    path_to_config = args.path_to_config
    path_to_data = args.path_to_data

    is_overwrite: bool = args.overwrite
    is_download: bool = args.no_download
    is_unzip: bool = args.no_unzip
    is_delete: bool = args.no_delete
    is_create_subdirs: bool = args.no_subdirs
    is_log: bool = args.no_log
    is_debug: bool = args.debug
    idx_selected_input: int | None = args.select_file_at_index
    idx_selected: int | None = parse_selected_index(args.select_file_at_index)
    is_export_results: bool = args.export_results

    product_type: list[str] = args.product_type
    product_version: str = args.product_version
    orbit_number: list[int] | None = args.orbit_number
    start_orbit_number: int | None = args.start_orbit_number
    end_orbit_number: int | None = args.end_orbit_number
    frame_id: list[str] | None = args.frame_id
    orbit_and_frame: list[str] | None = args.orbit_and_frame
    start_orbit_and_frame: str | None = args.start_orbit_and_frame
    end_orbit_and_frame: str | None = args.end_orbit_and_frame
    timestamps: list[str] | None = args.time
    start_time: str | None = args.start_time
    end_time: str | None = args.end_time
    radius_search: list[float] | None = args.radius_search
    bounding_box: list[float] | None = args.bounding_box

    ecdownload(
        file_type=product_type,
        product_version=product_version,
        orbit_number=orbit_number,
        start_orbit_number=start_orbit_number,
        end_orbit_number=end_orbit_number,
        frame_id=frame_id,
        orbit_and_frame=orbit_and_frame,
        start_orbit_and_frame=start_orbit_and_frame,
        end_orbit_and_frame=end_orbit_and_frame,
        timestamps=timestamps,
        start_time=start_time,
        end_time=end_time,
        radius_search=radius_search,
        bounding_box=bounding_box,
        path_to_config=path_to_config,
        path_to_data=path_to_data,
        is_log=is_log,
        is_debug=is_debug,
        is_download=is_download,
        is_overwrite=is_overwrite,
        is_unzip=is_unzip,
        is_delete=is_delete,
        is_create_subdirs=is_create_subdirs,
        is_export_results=is_export_results,
        idx_selected_input=idx_selected_input,
    )


def main() -> None:
    cli_tool_ecdownload()


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from logging import Logger
from typing import Literal, Sequence

import numpy as np
import xarray as xr
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from ..figure import ECKFigure


@dataclass
class _QuicklookResults:
    fig: Figure
    subfigs: list[list[ECKFigure]]



import os
from logging import Logger
from zipfile import BadZipFile, ZipFile

from ..utils._cli import console_exclusive_info, get_counter_message


def unzip_file(
    filepath: str,
    delete: bool = False,
    delete_on_error: bool = False,
    counter: int | None = None,
    total_count: int | None = None,
    logger: Logger | None = None,
) -> bool:
    """
    Extracts file and optionally deletes the original ZIP file upon success or error.

    Args:
        filepath (str): The path to the ZIP file to be extracted.
        delete (bool, optional): If True, the original ZIP file is deleted after extraction. Defaults to False.
        delete_on_error (bool, optional): If True, the ZIP file is deleted if an error occurs during extraction. Defaults to False.
        counter (int or None, optional): A counter to track progress during extraction. Defaults to None.
        total_count (int or None, optional): The total number of files to extract, used for progress tracking. Defaults to None.
        logger (Logger or None, optional): A logger instance to log progress and errors. Defaults to None.

    Returns:
        bool: True if the extraction was successful, False otherwise.
    """
    count_msg, _ = get_counter_message(counter=counter, total_count=total_count)

    if not os.path.exists(filepath):
        if logger:
            logger.info(f" {count_msg} File not found: <{filepath}>")
        return False

    if logger:
        console_exclusive_info(f" {count_msg} Extracting...", end="\r")
    new_filepath = os.path.join(
        os.path.dirname(filepath), os.path.basename(filepath).split(".")[0]
    )
    try:
        with ZipFile(filepath, "r") as zip_file:
            zip_file.extractall(path=new_filepath)
    except BadZipFile as e:
        if delete_on_error:
            os.remove(filepath)
            if logger:
                logger.info(f" {count_msg} Unzip failed! ZIP-file was deleted.")
        else:
            if logger:
                logger.info(f" {count_msg} Unzip failed! <{filepath}>")
        return False

    if delete:
        os.remove(filepath)
        if logger:
            logger.info(
                f" {count_msg} File extracted and ZIP-file deleted. (see <{new_filepath}>)"
            )
    else:
        if logger:
            logger.info(f" {count_msg} File extracted. (see <{new_filepath}>)")

    return True



import numpy as np
from numpy.typing import ArrayLike


def all_in(subset: ArrayLike, set: ArrayLike) -> bool:
    """
    Check if all elements in `subset` are present in `set`.

    Args:
        subset (ArrayLike): The list to check.
        set (ArrayLike): The list to check against.

    Returns:
        bool: True if all elements of `subset` are in `set`, False otherwise.
    """
    return all(item in np.asarray(set) for item in np.asarray(subset))



import xarray as xr

from ....constants import (
    DEFAULT_READ_EC_PRODUCT_HEADER,
    DEFAULT_READ_EC_PRODUCT_META,
    DEFAULT_READ_EC_PRODUCT_MODIFY,
)
from ....swath_data.across_track_distance import (
    add_across_track_distance,
    add_nadir_track,
    add_nadir_var,
    get_nadir_index,
)
from ....xarray_utils import merge_datasets
from .._rename_dataset_content import rename_common_dims_and_vars, rename_var_info
from ..file_info import FileAgency
from ..header_group import add_header_and_meta_data
from ..science_group import read_science_data


def read_product_amcth(
    filepath: str,
    modify: bool = DEFAULT_READ_EC_PRODUCT_MODIFY,
    header: bool = DEFAULT_READ_EC_PRODUCT_HEADER,
    meta: bool = DEFAULT_READ_EC_PRODUCT_META,
) -> xr.Dataset:
    """Opens AM__CTH_2A file as a `xarray.Dataset`."""
    ds = read_science_data(filepath, agency=FileAgency.ESA)

    if not modify:
        return ds

    nadir_idx = get_nadir_index(ds, nadir_idx=150)

    ds = ds.rename({"latitude": "swath_latitude"})
    ds = ds.rename({"longitude": "swath_longitude"})
    ds = add_nadir_track(
        ds,
        nadir_idx,
        swath_lat_var="swath_latitude",
        swath_lon_var="swath_longitude",
        along_track_dim="along_track",
        across_track_dim="across_track",
        nadir_lat_var="latitude",
        nadir_lon_var="longitude",
    )
    ds = add_across_track_distance(
        ds, nadir_idx, swath_lat_var="swath_latitude", swath_lon_var="swath_longitude"
    )

    ds = rename_common_dims_and_vars(
        ds,
        along_track_dim="along_track",
        across_track_dim="across_track",
        track_lat_var="latitude",
        track_lon_var="longitude",
        swath_lat_var="swath_latitude",
        swath_lon_var="swath_longitude",
        time_var="time",
    )

    ds["cloud_top_height_MSI"].values = ds["cloud_top_height_MSI"].values / 1000.0
    ds = rename_var_info(
        ds, "cloud_top_height_MSI", long_name="CTH from M-COP", units="km"
    )

    ds["cloud_top_height_difference_ATLID_MSI"].values = (
        ds["cloud_top_height_difference_ATLID_MSI"].values / 1000.0
    )
    ds = rename_var_info(
        ds,
        "cloud_top_height_difference_ATLID_MSI",
        long_name="AM-CTH CTH difference",
        units="km",
    )

    ds = rename_var_info(
        ds,
        "quality_status",
        long_name="Quality status",
        units="",
    )

    ds["plot_cloud_top_height_difference_ATLID_MSI"] = ds[
        "cloud_top_height_difference_ATLID_MSI"
    ].copy()
    mask = ds["quality_status"].values == 4
    ds["plot_cloud_top_height_difference_ATLID_MSI"].values[mask] = 100e3

    ds = add_nadir_var(ds, "cloud_top_height_MSI")
    ds = add_nadir_var(ds, "cloud_top_height_difference_ATLID_MSI")
    ds = add_nadir_var(ds, "quality_status")
    ds = add_nadir_var(ds, "cloud_fraction")

    ds = add_header_and_meta_data(filepath=filepath, ds=ds, header=header, meta=meta)

    return ds



from .agency import FileAgency
from .baseline import validate_baseline
from .file_info import FileInfoEnum
from .latency import FileLatency
from .mission_id import FileMissionID
from .orbit_and_frame import (
    format_frame_id,
    format_orbit_and_frame,
    format_orbit_number,
)
from .product_info import (
    ProductDataFrame,
    ProductInfo,
    get_product_info,
    get_product_infos,
    is_earthcare_product,
)
from .type import FileType, get_file_type



from typing import Iterable, Literal

import numpy as np
from numpy.typing import ArrayLike, NDArray
from scipy.interpolate import interp1d  # type: ignore

from ..geo import haversine, interpgeo
from ..statistics import nan_mean
from ..time import TimestampLike, num_to_time, time_to_num
from ._validate_dimensions import ensure_vertical_2d, validate_profile_data_dimensions


def _convert_height_centers_to_bins(height: ArrayLike) -> NDArray:
    height = np.asarray(height)

    if len(height.shape) != 1:
        raise ValueError(f"height is {len(height.shape)}D but exptected to be 1D.")

    hd1 = np.diff(height)
    hd2 = np.append(hd1[0], hd1)
    hd3 = np.append(hd1, hd1[-1])

    hnew1 = height - hd2 / 2
    hnew2 = height + hd3 / 2

    hnew = np.append(hnew1, hnew2[-1])
    return hnew


def rebin_height(
    values: Iterable[float] | NDArray,
    height: Iterable[float] | NDArray,
    new_height: Iterable[float] | NDArray,
    method: Literal["interpolate", "mean"] = "mean",
) -> NDArray:
    """
    Rebins profile data to new height bins.

    Parameters:
        values (np.ndarray):
            Profile values as a 2D array
            (shape represents temporal and vertical dimension).
        height (np.ndarray):
            Height values either as a 2D array (same dimensions as `values`)
            or as a 1D array (shape represents vertical dimension).
        new_height (np.ndarray):
            Target height bin centers as a 1D array (shape represents vertical dimension)

    Returns:
        rebinned_values (np.ndarray):
            2D array with values rebinned along the second (i.e. vertical) according to `new_height`.
    """
    values = np.asarray(values)
    height = np.asarray(height)
    new_height = np.asarray(new_height)

    validate_profile_data_dimensions(values, height=height)
    if len(new_height.shape) == 2 and new_height.shape[0] == 1:
        new_height = np.asarray(new_height[0])

    if len(new_height.shape) != 1:
        raise ValueError(
            f"Target height bins must be 1D but has {len(new_height.shape)} dimensions (shape={new_height.shape})"
        )

    M, _ = values.shape
    H = len(new_height)
    rebinned_values = np.full((M, H), np.nan)

    height = ensure_vertical_2d(height, M)
    for i in range(M):
        valid = np.isfinite(values[i])
        if np.sum(valid) > 1:
            try:
                if method == "interpolate":
                    interp = interp1d(
                        height[i, valid],
                        values[i, valid],
                        kind="linear",
                        bounds_error=False,
                        fill_value=np.nan,
                        assume_sorted=True,
                    )
                    rebinned_values[i] = interp(new_height)
                else:
                    bin_edges = _convert_height_centers_to_bins(new_height)
                    start_idxs = np.searchsorted(height[i], bin_edges[:-1], side="left")
                    end_idxs = np.searchsorted(height[i], bin_edges[1:], side="left")
                    rebinned_values[i] = np.array(
                        [
                            nan_mean(values[i, start:end]) if start < end else np.nan
                            for start, end in zip(start_idxs, end_idxs)
                        ]
                    )

            except Exception:
                continue

    return rebinned_values


def rebin_time(
    values: ArrayLike,
    time: ArrayLike,
    new_time: ArrayLike,
    is_geo: bool = False,
    method: Literal["interpolate", "mean"] = "mean",
) -> NDArray:
    """
    Rebins profile data to new time bins. If `is_geographic` is True, performs geodesic interpolation
    appropriate for latitude and longitude data.

    Args:
        values (np.ndarray): 2D array of values, shape (T, N).
        time (np.ndarray): 1D array of times (datetime64).
        new_time (np.ndarray): 1D array of target times (datetime64).
        is_geographic (bool): If True, apply geodesic interpolation for lon/lat.

    Returns:
        np.ndarray: Rebinned values with shape (len(new_time), N).
    """
    values = np.asarray(values)
    time = np.asarray(time)
    new_time = np.asarray(new_time)

    validate_profile_data_dimensions(values, time=time)

    ref_time = time[0].astype("datetime64[ns]")
    # print("ref_time", ref_time)
    time_f = (time - ref_time).astype("timedelta64[ns]").astype(np.float64)
    # print("time_f", time_f)
    new_time_f = (new_time - ref_time).astype("timedelta64[ns]").astype(np.float64)
    # print("new_time_f", new_time_f)

    T, N = len(new_time), values.shape[1]
    if is_geo:
        rebinned = np.full((T, 2), np.nan)
    else:
        rebinned = np.full((T, N), np.nan)

    for i in range(N):
        valid = np.isfinite(values[:, i])
        if np.sum(valid) < 2:
            continue

        if is_geo:
            # interpolate each dimension separately (lat and lon)
            for j, t_new in enumerate(new_time_f):
                # Find surrounding time points
                mask = time_f[valid] <= t_new
                if np.all(~mask) or np.all(mask):
                    continue
                idx_before = np.max(np.where(mask)[0])
                idx_after = np.min(np.where(~mask)[0])
                # print(time_f)
                t0, t1 = time_f[valid][[idx_before, idx_after]]
                # print(idx_before, idx_after)
                # print(t0, t1, values.shape)
                v0, v1 = values[valid][[idx_before, idx_after]]
                # print(v0, v1)
                f = (t_new - t0) / (t1 - t0)

                lon0, lat0 = v0
                lon1, lat1 = v1
                lon, lat = interpgeo(lon0, lat0, lon1, lat1, f)
                rebinned[j] = np.array(
                    [
                        lon,
                        lat,
                    ]
                )
        else:
            try:
                if method == "interpolate":
                    interp = interp1d(
                        time_f[valid],
                        values[valid, i],
                        kind="linear",
                        bounds_error=False,
                        fill_value=np.nan,
                    )
                    rebinned[:, i] = interp(new_time_f)
                else:
                    bin_edges = _convert_height_centers_to_bins(new_time_f)
                    start_idxs = np.searchsorted(
                        time_f[valid], bin_edges[:-1], side="left"
                    )
                    end_idxs = np.searchsorted(
                        time_f[valid], bin_edges[1:], side="left"
                    )
                    rebinned[:, i] = np.array(
                        [
                            nan_mean(values[start:end, i]) if start < end else np.nan
                            for start, end in zip(start_idxs, end_idxs)
                        ]
                    )

            except Exception:
                continue

    return rebinned


def rebin_along_track(
    values: ArrayLike,
    lat: ArrayLike,
    lon: ArrayLike,
    lat2: ArrayLike,
    lon2: ArrayLike,
) -> NDArray:
    """
    Interpolates values along track coordinates defined by lat/lon
    onto a new track's coordinates defined by lat2/lon2.

    Args:
        values (ArrayLike of shape (n,) or (n, m)): Values along the original track.
        lat (ArrayLike of shape (n,)): Original latitude.
        lon (ArrayLike of shape (n,)): Original longitude.
        lat2 (ArrayLike of shape (k,)): New latitude to interpolate to.
        lon2 (ArrayLike of shape (k,)): New longitude to interpolate to.

    Returns
        NDArray of shape (k,) or (k, m): Interpolated values at new track coordniates.
    """
    input = np.asarray(values)
    lat = np.asarray(lat)
    lon = np.asarray(lon)
    lat2 = np.asarray(lat2)
    lon2 = np.asarray(lon2)

    is_time: bool = isinstance(input[0], TimestampLike)

    if is_time:
        values = time_to_num(input, input[0])
    else:
        values = input

    # Calculate cumulative distances along track 1
    dists = haversine(
        np.vstack((lat[:-1], lon[:-1])).T,
        np.vstack((lat[1:], lon[1:])).T,
    )
    cum_dists = np.insert(np.cumsum(dists), 0, 0)

    # Calculate cumulative distances along track 2
    dists2 = haversine(
        np.vstack((lat2[:-1], lon2[:-1])).T,
        np.vstack((lat2[1:], lon2[1:])).T,
    )
    cum_dists2 = np.insert(np.cumsum(dists2), 0, 0)

    # Interpolate at points of track 2
    result = None
    if values.ndim == 1:
        result = np.interp(cum_dists2, cum_dists, values)
    elif values.ndim == 2:
        result = np.vstack(
            [
                np.interp(cum_dists2, cum_dists, values[:, i])
                for i in range(values.shape[1])
            ]
        ).T
    else:
        raise ValueError("values must be 1D or 2D array")

    if is_time:
        result = num_to_time(result, input[0])

    return np.asarray(result)



from .atl_aer_2a import read_product_aaer
from .atl_ald_2a import read_product_aald
from .atl_atc_2a import read_product_atc
from .atl_cla_2a import read_product_acla
from .atl_cth_2a import read_product_acth
from .atl_ebd_2a import read_product_aebd



import numpy as np
from numpy.typing import NDArray


def validate_profile_data_dimensions(
    values: NDArray,
    time: NDArray | None = None,
    height: NDArray | None = None,
    latitude: NDArray | None = None,
    longitude: NDArray | None = None,
) -> None:
    if len(values.shape) != 2:
        raise ValueError(
            f"Values must be either 2D, but has {len(values.shape)} dimensions (shape={values.shape})"
        )

    if isinstance(height, np.ndarray):
        is_height_1d = len(height.shape) == 1
        is_height_2d = len(height.shape) == 2
        if not is_height_1d and not is_height_2d:
            raise ValueError(
                f"Height must be either 1D or 2D, but has {len(height.shape)} dimensions (shape={height.shape})"
            )

        if is_height_1d:
            if height.shape[0] != values.shape[1]:
                raise ValueError(
                    f"Since height given is 1D it must have the same length as the second (i.e. vertical) dimension of values [height{height.shape[0]} != values{values.shape[1]}]"
                )
        elif is_height_2d:
            if height.shape != values.shape:
                raise ValueError(
                    f"Since height given is 2D it must have the same shape as values [height{height.shape} != values{values.shape}]"
                )

    if isinstance(time, np.ndarray):
        if time.shape[0] != values.shape[0]:
            raise ValueError(
                f"Time must have the same length as the first (i.e. temporal) dimension of values [time{time.shape} != values{values.shape}]"
            )

    if isinstance(latitude, np.ndarray):
        if len(latitude.shape) != 0 and latitude.shape[0] != values.shape[0]:
            raise ValueError(
                f"Latitude must have the same length as the first (i.e. temporal) dimension of values [latitude{latitude.shape} != values{values.shape}]"
            )

    if isinstance(longitude, np.ndarray):
        if len(longitude.shape) != 0 and longitude.shape[0] != values.shape[0]:
            raise ValueError(
                f"Longitude must have the same length as the first (i.e. temporal) dimension of values [longitude{longitude.shape} != values{values.shape}]"
            )


def ensure_vertical_2d(values: NDArray, vertical_dim_length: int) -> NDArray:
    if len(values.shape) == 1:
        values = np.repeat(values[np.newaxis, :], vertical_dim_length, axis=0)
    return values


def ensure_along_track_2d(values: NDArray, along_track_dim_length: int) -> NDArray:
    if len(values.shape) == 1:
        values = np.repeat(values[:, np.newaxis], along_track_dim_length, axis=1)
    return values



import os
from typing import overload

import xarray as xr

from ..header_group import read_header_data
from .file_info import FileInfoEnum


class FileAgency(FileInfoEnum):
    ESA = "E"
    JAXA = "J"

    @classmethod
    def from_input(cls, input: str | xr.Dataset) -> "FileAgency":
        """Infers the EarthCARE product agency (i.e. ESA or JAXA) from a given file or dataset."""
        if isinstance(input, str):
            try:
                return cls[input.upper()]
            except AttributeError:
                pass
            except KeyError:
                pass
            try:
                return cls(input.upper())
            except ValueError:
                pass

        return get_file_agency(input)


def _get_file_agency_from_dataset(ds: xr.Dataset) -> FileAgency:
    try:
        return FileAgency(str(ds.File_Class.values)[0])
    except AttributeError as e:
        filepath = ds.encoding["source"]
        filename = os.path.basename(filepath)
        file_class = filename.split(".")[0].split("_")[1]
        return FileAgency(file_class[0])


@overload
def get_file_agency(product: str) -> FileAgency: ...
@overload
def get_file_agency(product: xr.Dataset) -> FileAgency: ...
def get_file_agency(product: str | xr.Dataset) -> FileAgency:
    if isinstance(product, str):
        with read_header_data(product) as ds:
            file_class = _get_file_agency_from_dataset(ds)
    elif isinstance(product, xr.Dataset):
        file_class = _get_file_agency_from_dataset(product)
    else:
        raise NotImplementedError()
    return file_class



import json
import urllib.parse as urlp
from logging import Logger

import requests


def _encode_url(url: str | bytes) -> str | bytes:
    """Encode the URL, including its query string."""
    if isinstance(url, bytes):
        return url

    split_parsed_url = urlp.urlsplit(url)
    encoded_query = urlp.quote(split_parsed_url.query, safe="=&,/:")  # Keep separators
    return urlp.urlunsplit(
        (
            split_parsed_url.scheme,
            split_parsed_url.netloc,
            split_parsed_url.path,
            encoded_query,
            split_parsed_url.fragment,
        )
    )


def validate_request_response(
    response: requests.models.Response,
    logger: Logger | None = None,
) -> None:
    """Raises HTTPError if one occurred and logs it."""
    try:
        response.raise_for_status()
    except requests.HTTPError as e:
        if logger:
            logger.exception(e)
        raise


def _load_response_json(response: requests.Response) -> dict:
    """Reads a request response in JSON format."""
    return json.loads(response.text)


def request_get(
    url: str | bytes,
    logger: Logger | None = None,
    **kwargs,
) -> requests.Response:
    """Sends a GET request, validates it's response and returns it."""
    url = _encode_url(url=url)
    if logger:
        logger.debug(f"Send GET request: {str(url)}")
    response = requests.get(url, **kwargs)
    validate_request_response(response=response, logger=logger)
    return response


def get_request_json(url: str, logger: Logger | None = None, **kwargs) -> dict:
    response = request_get(url=url, logger=logger, **kwargs)
    data = _load_response_json(response)
    return data


def request_post(
    url: str | bytes,
    logger: Logger | None = None,
    **kwargs,
) -> requests.Response:
    """Sends a POST request, validates it's response and returns it."""
    url = _encode_url(url=url)
    if logger:
        logger.debug(f"Send POST request: {str(url)}")
    response = requests.post(url, **kwargs)
    validate_request_response(response=response, logger=logger)
    return response



import xarray as xr

from ....constants import (
    DEFAULT_READ_EC_PRODUCT_HEADER,
    DEFAULT_READ_EC_PRODUCT_META,
    DEFAULT_READ_EC_PRODUCT_MODIFY,
)
from ....xarray_utils import merge_datasets
from .._rename_dataset_content import (
    BSC_LABEL,
    DEPOL_LABEL,
    ELEVATION_VAR,
    EXT_LABEL,
    LR_LABEL,
    TROPOPAUSE_VAR,
    rename_common_dims_and_vars,
    rename_var_info,
)
from ..file_info import FileAgency
from ..header_group import add_header_and_meta_data
from ..science_group import read_science_data


def read_product_aebd(
    filepath: str,
    modify: bool = DEFAULT_READ_EC_PRODUCT_MODIFY,
    header: bool = DEFAULT_READ_EC_PRODUCT_HEADER,
    meta: bool = DEFAULT_READ_EC_PRODUCT_META,
) -> xr.Dataset:
    """Opens ATL_EBD_2A file as a `xarray.Dataset`."""
    ds = read_science_data(filepath, agency=FileAgency.ESA)

    if not modify:
        return ds

    ds = rename_common_dims_and_vars(
        ds,
        along_track_dim="along_track",
        vertical_dim="JSG_height",
        track_lat_var="latitude",
        track_lon_var="longitude",
        height_var="height",
        time_var="time",
        tropopause_var="tropopause_height",
        elevation_var="elevation",
    )

    bsc_long_name = lambda resolution: f"{BSC_LABEL} ({resolution})"
    ext_long_name = lambda resolution: f"{EXT_LABEL} ({resolution})"
    lr_long_name = lambda resolution: f"{LR_LABEL} ({resolution})"
    depol_long_name = lambda resolution: f"{DEPOL_LABEL} ({resolution})"

    for res, res_label in zip(
        ["", "_medium_resolution", "_low_resolution"], ["HiRes", "MedRes", "LowRes"]
    ):
        ds = rename_var_info(
            ds,
            f"particle_backscatter_coefficient_355nm{res}",
            name=bsc_long_name(res_label),
            long_name=bsc_long_name(res_label),
        )
        ds = rename_var_info(
            ds,
            f"particle_extinction_coefficient_355nm{res}",
            name=ext_long_name(res_label),
            long_name=ext_long_name(res_label),
        )
        ds = rename_var_info(
            ds,
            f"lidar_ratio_355nm{res}",
            name=lr_long_name(res_label),
            long_name=lr_long_name(res_label),
        )
        ds = rename_var_info(
            ds,
            f"particle_linear_depol_ratio_355nm{res}",
            name=depol_long_name(res_label),
            long_name=depol_long_name(res_label),
            units="-",
        )

    ds = add_header_and_meta_data(filepath=filepath, ds=ds, header=header, meta=meta)

    return ds



import xarray as xr

from ....constants import (
    DEFAULT_READ_EC_PRODUCT_HEADER,
    DEFAULT_READ_EC_PRODUCT_META,
    DEFAULT_READ_EC_PRODUCT_MODIFY,
)
from ....xarray_utils import merge_datasets
from .._rename_dataset_content import rename_common_dims_and_vars
from ..file_info import FileAgency
from ..header_group import add_header_and_meta_data
from ..science_group import read_science_data


def read_product_aald(
    filepath: str,
    modify: bool = DEFAULT_READ_EC_PRODUCT_MODIFY,
    header: bool = DEFAULT_READ_EC_PRODUCT_HEADER,
    meta: bool = DEFAULT_READ_EC_PRODUCT_META,
) -> xr.Dataset:
    """Opens ATL_ALD_2A file as a `xarray.Dataset`."""
    ds = read_science_data(filepath, agency=FileAgency.ESA)

    if not modify:
        return ds

    ds = rename_common_dims_and_vars(
        ds,
        along_track_dim="along_track",
        track_lat_var="latitude",
        track_lon_var="longitude",
        time_var="time",
    )

    vars = [
        "aerosol_layer_top",
        "aerosol_layer_confidence",
        "aerosol_layer_base_confidence",
        "aerosol_layer_top_confidence",
        "aerosol_layer_optical_thickness_355nm",
        "aerosol_layer_optical_thickness_355nm_error",
        "aerosol_layer_mean_extinction_355nm",
        "aerosol_layer_mean_extinction_355nm_error",
        "aerosol_layer_mean_backscatter_355nm",
        "aerosol_layer_mean_backscatter_355nm_error",
        "aerosol_layer_mean_lidar_ratio_355nm",
        "aerosol_layer_mean_lidar_ratio_355nm_error",
        "aerosol_layer_mean_depolarisation_355nm",
        "aerosol_layer_mean_depolarisation_355nm_error",
    ]

    n_layers = ds["max_layers"].shape[0]

    for i in range(n_layers):
        for var in vars:
            ds = ds.assign({f"{var}{i+1}": ds[var].isel({"max_layers": i})})

    ds = add_header_and_meta_data(filepath=filepath, ds=ds, header=header, meta=meta)

    return ds



from enum import IntEnum
from typing import TypeAlias

from .curtain import CurtainFigure
from .map import MapFigure
from .profile import ProfileFigure
from .swath import SwathFigure

ECKFigure: TypeAlias = MapFigure | CurtainFigure | SwathFigure | ProfileFigure


class FigureType(IntEnum):
    NONE = -1
    CURTAIN = 0
    CURTAIN_ZOOMED = 1
    SWATH = 2
    SWATH_ZOOMED = 3
    MAP_1_ROW = 4
    MAP_2_ROW = 5
    PROFILE = 6
    LINE = 7
    LINE_ZOOMED = 8



import xarray as xr

from ....constants import (
    DEFAULT_READ_EC_PRODUCT_HEADER,
    DEFAULT_READ_EC_PRODUCT_META,
    DEFAULT_READ_EC_PRODUCT_MODIFY,
)
from ....xarray_utils import merge_datasets
from .._rename_dataset_content import rename_common_dims_and_vars
from ..file_info import FileAgency
from ..header_group import add_header_and_meta_data
from ..science_group import read_science_data


def read_product_acth(
    filepath: str,
    modify: bool = DEFAULT_READ_EC_PRODUCT_MODIFY,
    header: bool = DEFAULT_READ_EC_PRODUCT_HEADER,
    meta: bool = DEFAULT_READ_EC_PRODUCT_META,
) -> xr.Dataset:
    """Opens ATL_CTH_2A file as a `xarray.Dataset`."""
    ds = read_science_data(filepath, agency=FileAgency.ESA)

    if not modify:
        return ds

    ds = rename_common_dims_and_vars(
        ds,
        along_track_dim="along_track",
        track_lat_var="latitude",
        track_lon_var="longitude",
        time_var="time",
    )

    ds = add_header_and_meta_data(filepath=filepath, ds=ds, header=header, meta=meta)

    return ds



import warnings
from dataclasses import asdict, dataclass
from typing import Iterable, Literal, Sequence, Tuple, TypeAlias

import numpy as np
import pandas as pd
import xarray as xr
from numpy.typing import ArrayLike, NDArray

from .. import statistics as stats
from .._parse_units import parse_units
from ..constants import HEIGHT_VAR, TIME_VAR, TRACK_LAT_VAR, TRACK_LON_VAR
from ..geo import haversine
from ..np_array_utils import coarsen_mean, pad_true_sequence
from ..rolling_mean import rolling_mean_2d
from ..time import (
    TimeRangeLike,
    TimestampLike,
    num_to_time,
    time_to_num,
    to_timestamps,
    validate_time_range,
)
from ..typing import DistanceRangeLike
from ..validation import validate_height_range
from ._validate_dimensions import ensure_vertical_2d, validate_profile_data_dimensions
from .rebin import rebin_along_track, rebin_height, rebin_time


def _mean_2d(a: NDArray, axis: int = 0) -> NDArray:
    if len(a.shape) == 2:
        return np.array(nanmean(a, axis=axis))
    return np.array(a)


def _mean_1d(a: NDArray) -> NDArray:
    if a.dtype not in ["datetime64[ns]", "datetime64[s]"]:
        return np.array(nanmean(a))
    else:
        time = a
        reference_time = time[0].astype("datetime64[s]")
        time = (time - reference_time).astype("timedelta64[s]").astype(np.float64)
        new_time = np.array(nanmean(time))
        a = reference_time + new_time.astype("timedelta64[s]")
        return a


def nanmean(a: ArrayLike, axis: int | None = None) -> np.floating | NDArray:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        return np.nanmean(np.array(a), axis=axis)


@dataclass(frozen=True)
class ProfileStatResults:
    hmin: float
    hmax: float
    mean: float
    std: float
    mean_error: float | None

    def to_dict(self) -> dict:
        return asdict(self)

    def to_dataframe(self) -> pd.DataFrame:
        df = pd.DataFrame([self.to_dict()])
        df = df.astype(
            dict(
                hmin=float,
                hmax=float,
                mean=float,
                std=float,
                mean_error=float,
            )
        )
        return df


@dataclass(frozen=True)
class ProfileComparisonResults:
    hmin: float
    hmax: float
    diff_of_means: float
    mae: float
    rmse: float
    mean_diff: float
    prediction: ProfileStatResults
    target: ProfileStatResults

    def to_dict(self) -> dict:
        d = asdict(self)
        d_pred = d["prediction"].copy()
        d_targ = d["target"].copy()

        for k, v in d_pred.items():
            if k in ["hmin", "hmax"]:
                continue
            d[f"{k}_prediction"] = v

        for k, v in d_targ.items():
            if k in ["hmin", "hmax"]:
                continue
            d[f"{k}_target"] = v

        d = {k: v for k, v in d.items() if k not in ["prediction", "target"]}

        return d

    def to_dataframe(self) -> pd.DataFrame:
        df = pd.DataFrame([self.to_dict()])
        df = df.astype(
            dict(
                hmin=float,
                hmax=float,
                diff_of_means=float,
                mae=float,
                rmse=float,
                mean_diff=float,
                mean_prediction=float,
                std_prediction=float,
                mean_error_prediction=float,
                mean_target=float,
                std_target=float,
                mean_error_target=float,
            )
        )
        return df


@dataclass
class ProfileData:
    values: NDArray
    height: NDArray
    time: NDArray
    latitude: NDArray | None = None
    longitude: NDArray | None = None
    color: str | None = None
    label: str | None = None
    units: str | None = None
    platform: str | None = None
    error: NDArray | None = None

    def __post_init__(self):

        if isinstance(self.values, Iterable):
            self.values = np.atleast_2d(self.values)
        if isinstance(self.height, Iterable):
            self.height = np.asarray(self.height)
        if isinstance(self.time, Iterable):
            self.time = pd.to_datetime(np.asarray(self.time)).to_numpy()
        if isinstance(self.latitude, Iterable):
            self.latitude = np.asarray(self.latitude)
        if isinstance(self.longitude, Iterable):
            self.longitude = np.asarray(self.longitude)
        if isinstance(self.error, Iterable):
            self.error = np.atleast_2d(self.error)
            if self.values.shape != self.error.shape:
                raise ValueError(
                    f"`error` must have same shape as `values`: values.shape={self.values.shape} != error.shape={self.error.shape}"
                )
        if isinstance(self.units, str):
            self.units = parse_units(self.units)

        validate_profile_data_dimensions(
            values=self.values,
            height=self.height,
            time=self.time,
            latitude=self.latitude,
            longitude=self.longitude,
        )

    @property
    def shape(self):
        return self.values.shape

    @classmethod
    def from_dataset(
        self,
        ds: xr.Dataset,
        var: str,
        error_var: str | None = None,
        height_var: str = HEIGHT_VAR,
        time_var: str = TIME_VAR,
        lat_var: str = TRACK_LAT_VAR,
        lon_var: str = TRACK_LON_VAR,
        color: str | None = None,
        label: str | None = None,
        units: str | None = None,
        platform: str | None = None,
    ) -> "ProfileData":
        values = ds[var].values
        height = ds[height_var].values
        time = ds[time_var].values

        latitude: NDArray | None = None
        if lat_var in ds:
            latitude = ds[lat_var].values

        longitude: NDArray | None = None
        if lon_var in ds:
            longitude = ds[lon_var].values

        if not isinstance(label, str):
            label = None if not hasattr(ds[var], "long_name") else ds[var].long_name

        if not isinstance(label, str):
            label = None if not hasattr(ds[var], "name") else ds[var].name  # type: ignore

        if not isinstance(label, str):
            label = None if not hasattr(ds[var], "label") else ds[var].label

        if not isinstance(units, str):
            units = None if not hasattr(ds[var], "units") else ds[var].units

        if not isinstance(units, str):
            units = None if not hasattr(ds[var], "unit") else ds[var].unit

        error: NDArray | None = None
        if isinstance(error_var, str):
            error = ds[error_var].values

        return ProfileData(
            values=values,
            height=height,
            time=time,
            latitude=latitude,
            longitude=longitude,
            color=color,
            label=label,
            units=units,
            platform=platform,
            error=error,
        )

    def print_shapes(self):
        if isinstance(self.values, Iterable):
            print(f"values={self.values.shape}")
        if isinstance(self.height, Iterable):
            print(f"height={self.height.shape}")
        if isinstance(self.time, Iterable):
            print(f"time={self.time.shape}")
        if isinstance(self.latitude, Iterable):
            print(f"latitude={self.latitude.shape}")
        if isinstance(self.longitude, Iterable):
            print(f"longitude={self.longitude.shape}")

    def mean(self) -> "ProfileData":
        """Returns mean profile."""
        new_values = _mean_2d(self.values)
        new_height = _mean_2d(self.height)
        new_error: NDArray | None = None
        if isinstance(self.error, np.ndarray):
            new_error = _mean_2d(self.error)

        if isinstance(self.time, np.ndarray):
            new_time = _mean_1d(self.time)
        else:
            new_time = None

        if isinstance(self.latitude, np.ndarray):
            new_latitude = _mean_1d(self.latitude)
        else:
            new_latitude = None

        if isinstance(self.longitude, np.ndarray):
            new_longitude = _mean_1d(self.longitude)
        else:
            new_longitude = None

        new_color = self.color
        new_label = self.label
        new_units = self.units
        new_platform = self.platform

        return ProfileData(
            values=new_values,
            height=new_height,
            time=new_time,
            latitude=new_latitude,
            longitude=new_longitude,
            color=new_color,
            label=new_label,
            units=new_units,
            platform=new_platform,
            error=new_error,
        )

    def rolling_mean(self, window_size: int, axis: Literal[0, 1] = 0) -> "ProfileData":
        """Returns mean profile."""
        if len(self.values.shape) == 2:
            new_values = rolling_mean_2d(self.values, w=window_size, axis=axis)
            new_error: NDArray | None = None
            if isinstance(self.error, np.ndarray):
                new_error = self.error
            return ProfileData(
                values=new_values,
                height=self.height,
                time=self.time,
                latitude=self.latitude,
                longitude=self.longitude,
                color=self.color,
                label=self.label,
                units=self.units,
                platform=self.platform,
                error=new_error,
            )

        msg = f"VerticalProfile contains only one profile and thus {self.rolling_mean.__name__}() is not applied."
        warnings.warn(msg)
        return self

    def layer_mean(self, hmin: float, hmax: float) -> NDArray:
        """Returns layer mean values."""
        layer_mask = np.logical_and(hmin <= self.height, self.height <= hmax)
        layer_mean_values = self.values
        layer_mean_values[~layer_mask] = np.nan
        if len(layer_mean_values.shape) == 2:
            layer_mean_values = _mean_2d(layer_mean_values, axis=1)
        else:
            layer_mean_values = np.array(nanmean(layer_mean_values))
        return layer_mean_values

    def rebin_height(
        self,
        height_bin_centers: Iterable[float] | NDArray,
        method: Literal["interpolate", "mean"] = "mean",
    ) -> "ProfileData":
        """
        Rebins profiles to new height bins.

        Parameters:
            new_height (np.ndarray):
                Target height bin centers as a 1D array (shape represents vertical dimension)

        Returns:
            rebinned_profiles (VerticalProfiles):
                Profiles rebinned along the vertical dimension according to `height_bin_centers`.
        """
        if self.height.shape == np.array(height_bin_centers).shape and np.all(
            np.array(self.height) == np.array(height_bin_centers)
        ):
            return ProfileData(
                values=self.values,
                height=self.height,
                time=self.time,
                latitude=self.latitude,
                longitude=self.longitude,
                color=self.color,
                label=self.label,
                units=self.units,
                platform=self.platform,
                error=self.error,
            )

        new_values = rebin_height(
            self.values,
            self.height,
            height_bin_centers,
            method=method,
        )
        new_height = np.asarray(height_bin_centers)
        if len(new_values.shape) == 2:
            new_height = np.atleast_2d(new_height)
            if new_height.shape[0] == 1:
                new_height = new_height[0]
        new_error: NDArray | None = None
        if isinstance(self.error, np.ndarray):
            new_error = rebin_height(
                self.error,
                self.height,
                height_bin_centers,
                method=method,
            )
        return ProfileData(
            values=new_values,
            height=new_height,
            time=self.time,
            latitude=self.latitude,
            longitude=self.longitude,
            color=self.color,
            label=self.label,
            units=self.units,
            platform=self.platform,
            error=new_error,
        )

    def rebin_time(
        self,
        time_bin_centers: Sequence[TimestampLike] | ArrayLike,
        method: Literal["interpolate", "mean"] = "mean",
    ) -> "ProfileData":
        """
        Rebins profiles to new time bins.

        Args:
            time_bin_centers (Iterable[TimestampLike] | ArrayLike):
                Target time bin centers as a 1D array (shape represents temporal dimension)

        Returns:
            rebinned_profiles (VerticalProfiles):
                Profiles rebinned along the temporal dimension according to `height_bin_centers`.
        """
        time_bin_centers = to_timestamps(time_bin_centers)
        new_values = rebin_time(self.values, self.time, time_bin_centers, method=method)
        if len(self.height.shape) == 2:
            new_height = rebin_time(
                self.height, self.time, time_bin_centers, method=method
            )
        else:
            new_height = self.height
        new_error: NDArray | None = None
        if isinstance(self.error, np.ndarray):
            new_error = rebin_time(
                self.error, self.time, time_bin_centers, method=method
            )

        if isinstance(self.latitude, np.ndarray) and isinstance(
            self.longitude, np.ndarray
        ):
            new_coords = rebin_time(
                np.vstack([self.latitude, self.longitude]).T,
                self.time,
                time_bin_centers,
                is_geo=True,
                method=method,
            )
            new_latitude = new_coords[:, 0]
            new_longitude = new_coords[:, 0]
        else:
            new_latitude = None
            new_longitude = None
        return ProfileData(
            values=new_values,
            height=new_height,
            time=pd.to_datetime(to_timestamps(time_bin_centers)).to_numpy(),
            latitude=new_latitude,
            longitude=new_longitude,
            color=self.color,
            label=self.label,
            units=self.units,
            platform=self.platform,
            error=new_error,
        )

    def rebin_along_track(
        self,
        latitude_bin_centers: ArrayLike,
        longitude_bin_centers: ArrayLike,
    ) -> "ProfileData":
        """
        Rebins profiles to new time bins.

        Args:
            latitude_bin_centers (ArrayLike):
                Target time bin centers as a 1D array (shape represents temporal dimension)

        Returns:
            rebinned_profiles (VerticalProfiles):
                Profiles rebinned along the temporal dimension according to `height_bin_centers`.
        """
        has_lat = self.latitude is not None
        has_lon = self.longitude is not None

        if not has_lat or not has_lon:
            missing = []
            if not has_lat:
                missing.append("latitude")
            if not has_lon:
                missing.append("longitude")
            raise ValueError(
                f"{ProfileData.__name__} instance is missing {' and '.join(missing)} data"
            )

        latitude_bin_centers = np.asarray(latitude_bin_centers)
        longitude_bin_centers = np.asarray(longitude_bin_centers)

        new_values = rebin_along_track(
            self.values,
            np.asarray(self.latitude),
            np.asarray(self.longitude),
            latitude_bin_centers,
            longitude_bin_centers,
        )
        new_error: NDArray | None = None
        if isinstance(self.error, np.ndarray):
            new_error = rebin_along_track(
                self.error,
                np.asarray(self.latitude),
                np.asarray(self.longitude),
                latitude_bin_centers,
                longitude_bin_centers,
            )
        new_times = rebin_along_track(
            self.time,
            np.asarray(self.latitude),
            np.asarray(self.longitude),
            latitude_bin_centers,
            longitude_bin_centers,
        )
        return ProfileData(
            values=new_values,
            height=self.height,
            time=new_times,
            latitude=np.array(latitude_bin_centers),
            longitude=np.array(longitude_bin_centers),
            color=self.color,
            label=self.label,
            units=self.units,
            platform=self.platform,
            error=new_error,
        )

    def to_dict(self) -> dict:
        """Returns stored profile data as `dict`."""
        return asdict(self)

    def select_height_range(
        self,
        height_range: DistanceRangeLike,
    ) -> "ProfileData":
        """Retruns only data within the specified `height_range`."""
        height_range = validate_height_range(height_range)

        if len(self.height.shape) == 2:
            ref_height = self.height[0]
        else:
            ref_height = self.height

        mask = np.logical_and(
            height_range[0] <= ref_height, ref_height <= height_range[1]
        )

        sel_values = self.values[:, mask]
        sel_error: NDArray | None = None
        if isinstance(self.error, np.ndarray):
            sel_error = self.error[:, mask]

        if len(self.height.shape) == 2:
            sel_height = self.height[:, mask]
        else:
            sel_height = self.height[mask]

        return ProfileData(
            values=sel_values,
            height=sel_height,
            time=self.time,
            latitude=self.latitude,
            longitude=self.longitude,
            color=self.color,
            label=self.label,
            units=self.units,
            platform=self.platform,
            error=sel_error,
        )

    def select_time_range(
        self,
        time_range: TimeRangeLike | None,
        pad_idxs: int = 0,
    ) -> "ProfileData":
        """Retruns only data within the specified `time_range`."""
        if time_range is None:
            return self
        elif not isinstance(self.time, np.ndarray):
            raise ValueError(
                f"{ProfileData.__name__}.{self.select_time_range.__name__}() missing `time` data"
            )

        time_range = validate_time_range(time_range)

        times = to_timestamps(self.time)
        mask = np.logical_and(time_range[0] <= times, times <= time_range[1])
        mask = pad_true_sequence(mask, pad_idxs)

        sel_values = self.values[mask]
        sel_error: NDArray | None = None
        if isinstance(self.error, np.ndarray):
            sel_error = self.error[:, mask]
        sel_time = self.time[mask]

        if len(self.height.shape) == 2:
            sel_height = self.height[mask]
        else:
            sel_height = self.height

        if isinstance(self.latitude, np.ndarray):
            sel_latitude = self.latitude[mask]
        else:
            sel_latitude = None

        if isinstance(self.longitude, np.ndarray):
            sel_longitude = self.longitude[mask]
        else:
            sel_longitude = None

        return ProfileData(
            values=sel_values,
            height=sel_height,
            time=sel_time,
            latitude=sel_latitude,
            longitude=sel_longitude,
            color=self.color,
            label=self.label,
            units=self.units,
            platform=self.platform,
            error=sel_error,
        )

    def coarsen_mean(self, n: int, is_bin: bool = False) -> "ProfileData":
        """Returns downsampled profile data."""
        if len(self.values.shape) == 2:
            new_values: NDArray
            new_values = coarsen_mean(self.values, n=n, is_bin=is_bin)
            new_error: NDArray | None = None
            if isinstance(self.error, np.ndarray):
                new_error = coarsen_mean(self.error, n=n, is_bin=is_bin)
            new_time: NDArray = coarsen_mean(self.time, n=n)

            new_height: NDArray
            if len(self.height.shape) == 2:
                new_height = coarsen_mean(self.height, n=n)
            else:
                new_height = self.height

            new_latitude: NDArray | None
            if isinstance(self.latitude, np.ndarray):
                new_latitude = coarsen_mean(self.latitude, n=n)
            else:
                new_latitude = None

            new_longitude: NDArray | None
            if isinstance(self.longitude, np.ndarray):
                new_longitude = coarsen_mean(self.longitude, n=n)
            else:
                new_longitude = None

            return ProfileData(
                values=new_values,
                height=new_height,
                time=new_time,
                latitude=new_latitude,
                longitude=new_longitude,
                color=self.color,
                label=self.label,
                units=self.units,
                platform=self.platform,
                error=new_error,
            )

        msg = f"VerticalProfile contains only one profile and thus {self.coarsen_mean.__name__}() is not applied."
        warnings.warn(msg)
        return self

    def stats(
        self,
        height_range: DistanceRangeLike | None = None,
    ) -> ProfileStatResults:
        p = self
        _hmin: float = float(np.nanmin(p.height))
        _hmax: float = float(np.nanmax(p.height))
        if height_range is not None:
            height_range = validate_height_range(height_range)
            _hmin = height_range[0]
            _hmax = height_range[1]
            p = p.select_height_range(height_range)

        p = p.mean()
        _mean: float = float(stats.nan_mean(p.values))
        _std: float = float(stats.nan_std(p.values))
        _mean_error: float | None = None
        if isinstance(p.error, np.ndarray):
            _mean_error = float(stats.nan_mean(p.error))
        return ProfileStatResults(
            hmin=_hmin,
            hmax=_hmax,
            mean=_mean,
            std=_std,
            mean_error=_mean_error,
        )

    def compare_to(
        self,
        target: "ProfileData",
        height_range: DistanceRangeLike | None = None,
    ) -> ProfileComparisonResults:
        p = self
        p = p.mean()
        t = target
        t = t.mean()

        get_mean_abs_diff = lambda x: float(np.nanmean(np.abs(np.diff(x))))
        if get_mean_abs_diff(p.height) > get_mean_abs_diff(t.height):
            t = t.rebin_height(p.height)
        else:
            p = p.rebin_height(t.height)

        _hmin: float = float(np.nanmin(p.height))
        _hmax: float = float(np.nanmax(p.height))
        if height_range is not None:
            height_range = validate_height_range(height_range)
            _hmin = height_range[0]
            _hmax = height_range[1]
            p = p.select_height_range(height_range)
            t = t.select_height_range(height_range)

        stats_pred = p.stats()
        stats_targ = t.stats()
        _diff_of_means: float = float(stats.nan_diff_of_means(p.values, t.values))
        _mae: float = float(stats.nan_mae(p.values, t.values))
        _rmse: float = float(stats.nan_rmse(p.values, t.values))
        _mean_diff: float = float(stats.nan_mean_diff(p.values, t.values))

        return ProfileComparisonResults(
            hmin=_hmin,
            hmax=_hmax,
            diff_of_means=_diff_of_means,
            mae=_mae,
            rmse=_rmse,
            mean_diff=_mean_diff,
            prediction=stats_pred,
            target=stats_targ,
        )

    def to_mega(self) -> "ProfileData":
        import logging

        logger = logging.getLogger()

        if isinstance(self.units, str):
            if self.units in ["m-1 sr-1", "m-1"]:
                return ProfileData(
                    values=self.values * 1e6,
                    height=self.height,
                    time=self.time,
                    latitude=self.latitude,
                    longitude=self.longitude,
                    color=self.color,
                    label=self.label,
                    units=f"M{self.units}",
                    platform=self.platform,
                    error=(
                        None
                        if not isinstance(self.error, np.ndarray)
                        else self.error * 1e6
                    ),
                )
            elif self.units in ["Mm-1 sr-1", "Mm-1"]:
                logger.warning(
                    f"""Profile units already converted to "{self.units}"."""
                )
                return self.copy()
            else:
                logger.warning(
                    f"""Can not convert profile to "Mm-1 sr-1" or "Mm-1" since it's original units are: "{self.units}"."""
                )
                return self.copy()
        logger.warning(
            f"""Can not convert profile to "Mm-1 sr-1" or "Mm-1" since units are not given."""
        )

        return self.copy()

    def copy(self) -> "ProfileData":
        return ProfileData(
            values=self.values,
            height=self.height,
            time=self.time,
            latitude=self.latitude,
            longitude=self.longitude,
            color=self.color,
            label=self.label,
            units=self.units,
            platform=self.platform,
            error=self.error,
        )



import re

from ..utils._cli._parse import (
    parse_path_to_config,
    parse_path_to_data,
    parse_search_inputs,
    parse_selected_index,
)
from ._types import Entrypoint, UserType


def get_collection_names_available_to_user(
    user_type: UserType,
    entrypoint: Entrypoint,
) -> list[str]:
    colls_comm = [
        "EarthCAREL0L1Products",
        "EarthCAREL2Products",
        "JAXAL2Products",
    ]

    colls_calval = [
        "EarthCAREAuxiliary",
        "EarthCAREL1InstChecked",
        "EarthCAREL2InstChecked",
        "JAXAL2InstChecked",
    ]

    colls_public = [
        "EarthCAREL1Validated",
        "EarthCAREL2Validated",
        "JAXAL2Validated",
        "EarthCAREXMETL1DProducts10",
        "EarthCAREOrbitData",
    ]

    is_maap: bool = entrypoint == Entrypoint.MAAP
    is_oads: bool = not is_maap

    to_maap = lambda colls: [f"{c}_MAAP" for c in colls]
    if is_maap and is_oads:
        colls_comm = colls_comm + to_maap(colls_comm)
        colls_calval = colls_calval + to_maap(colls_calval)
        colls_public = colls_public + to_maap(colls_public)
    elif is_maap:
        colls_comm = to_maap(colls_comm)
        colls_calval = to_maap(colls_calval)
        colls_public = to_maap(colls_public)

    if user_type == UserType.COMMISSIONING:
        return colls_comm + colls_calval + colls_public
    elif user_type == UserType.CALVAL:
        return colls_calval + colls_public
    elif user_type == UserType.OPEN:
        return colls_public
    else:
        raise NotImplementedError()


def parse_user_type(user_type: str) -> UserType:
    if not isinstance(user_type, str):
        raise TypeError(
            f"invalid type '{type(user_type).__name__}' for user_type, expects 'str'"
        )

    user_type = re.sub(r"\W+", "", user_type.lower())
    user_type = user_type.replace("teams", "").replace("team", "").replace("user", "")
    if user_type == "commissioning":
        return UserType.COMMISSIONING
    elif user_type == "calval":
        return UserType.CALVAL
    elif user_type == "open":
        return UserType.OPEN
    else:
        raise ValueError(
            f'invalid input for user_type: "{user_type}". expected inputs are: "commissioning", "calval" or "open"'
        )



import logging

logger: logging.Logger = logging.getLogger(__name__)

from dataclasses import dataclass
from typing import Any, Literal, overload

import numpy as np
import pandas as pd
import xarray as xr
from numpy.typing import NDArray

from .config import read_config
from .constants import ALONG_TRACK_DIM, TIME_VAR, TRACK_LAT_VAR, TRACK_LON_VAR
from .geo import geodesic, get_coords, get_cumulative_distances
from .geo.string_formatting import format_coords
from .ground_sites import GroundSite, get_ground_site
from .np_array_utils import ismonotonic
from .read import read_product, search_product
from .time import TimestampLike, to_timestamp
from .typing import LatLonCoordsLike, validate_numeric_pair
from .xarray_utils import EmptyFilterResultError, filter_radius


@dataclass
class OverpassInfo:
    site_name: str
    site_lat_deg_north: float
    site_lon_deg_east: float
    site_radius_km: float
    start_index: int
    end_index: int
    start_time: pd.Timestamp
    end_time: pd.Timestamp
    start_lat_deg_north: float
    start_lon_deg_east: float
    end_lat_deg_north: float
    end_lon_deg_east: float
    closest_index: int
    closest_lat_deg_north: float
    closest_lon_deg_east: float
    closest_time: pd.Timestamp
    closest_distance_km: float
    along_track_distance_km: float
    frame_crosses_pole: bool
    samples: int
    site: GroundSite

    @property
    def site_coords(self) -> tuple[float, float]:
        """Returns lat/lon coordinates of the overpassed site or center."""
        return self.site_lat_deg_north, self.site_lon_deg_east

    @property
    def index_range(self) -> tuple[int, int]:
        """Returns start and end indecies of the overpass."""
        return self.start_index, self.end_index

    @property
    def time_range(self) -> tuple[pd.Timestamp, pd.Timestamp]:
        """Returns start and end times of the overpass."""
        return self.start_time, self.end_time

    @property
    def start_coords(self) -> tuple[float, float]:
        """Returns lat/lon coordinates of the satellite at the start of the overpass."""
        return self.start_lat_deg_north, self.start_lon_deg_east

    @property
    def end_coords(self) -> tuple[float, float]:
        """Returns lat/lon coordinates of the satellite at the end of the overpass."""
        return self.end_lat_deg_north, self.end_lon_deg_east

    @property
    def closest_coords(self) -> tuple[float, float]:
        """Returns lat/lon coordinates of the satellite at the end of the overpass."""
        return self.closest_lat_deg_north, self.closest_lon_deg_east

    @property
    def duration(self) -> pd.Timedelta:
        return self.end_time - self.start_time


def get_closest_distance(
    ds: xr.Dataset,
    *,
    site_lat: float | int | None = None,
    site_lon: float | int | None = None,
    site_name: str | None = None,
    lat_var: str = TRACK_LAT_VAR,
    lon_var: str = TRACK_LON_VAR,
) -> float:
    if not isinstance(site_name, str) and not (
        isinstance(site_lat, (float, int)),
        isinstance(site_lon, (float, int)),
    ):
        raise TypeError(
            f"Missing arguments. At least either `site_name` or `site_lat` and `site_lon` must be given."
        )

    if isinstance(site_name, str):
        site = get_ground_site(site_name)
        if not isinstance(site_lat, (float, int)):
            site_lat = site.latitude
        if not isinstance(site_lon, (float, int)):
            site_lon = site.longitude

    assert isinstance(site_lat, (float, int))
    assert isinstance(site_lon, (float, int))

    site_lat = float(site_lat)
    site_lon = float(site_lon)
    site_coords = (site_lat, site_lon)

    # Closest sample
    along_track_coords = get_coords(ds, lat_var=lat_var, lon_var=lon_var)
    distances = geodesic(along_track_coords, site_coords, units="km")
    closest_distance = float(np.min(distances))

    return closest_distance


def _get_overpass_info(
    ds: xr.Dataset,
    site_radius_km: float | int,
    site: GroundSite | str,
    *,
    time_var: str = TIME_VAR,
    lat_var: str = TRACK_LAT_VAR,
    lon_var: str = TRACK_LON_VAR,
    along_track_dim: str = ALONG_TRACK_DIM,
) -> OverpassInfo:
    _site: GroundSite
    if isinstance(site, str):
        _site = get_ground_site(site)
    elif isinstance(site, GroundSite):
        _site = site
    else:
        raise TypeError(
            f"invalid type '{type(site).__name__}' for site, expected type 'GroundSite' or 'str'"
        )

    site_name: str | None = _site.long_name

    site_lat = _site.latitude
    site_lon = _site.longitude

    assert isinstance(site_lat, (float, int))
    assert isinstance(site_lon, (float, int))

    site_lat = float(site_lat)
    site_lon = float(site_lon)
    site_coords = (site_lat, site_lon)

    try:
        ds_filtered = filter_radius(
            ds,
            radius_km=site_radius_km,
            site=site,
            lat_var=lat_var,
            lon_var=lon_var,
            along_track_dim=along_track_dim,
        )
    except EmptyFilterResultError as e:
        raise ValueError(
            f"This is not a valid overpass. Track does not overlap radius area."
        )

    # Times
    original_time = ds[time_var].values
    time = ds_filtered[time_var].values
    start_time = time[0]
    end_time = time[-1]

    assert start_time <= end_time

    # Duration
    duration = to_timestamp(end_time) - to_timestamp(start_time)

    assert duration >= pd.Timedelta(0)

    # Indexes
    start_index = np.argmin(np.abs(original_time - start_time))
    end_index = np.argmin(np.abs(original_time - end_time))

    assert start_index <= end_index

    # Latitudes
    original_lat = ds[lat_var].values
    lat = ds_filtered[lat_var].values
    start_lat = lat[0]
    end_lat = lat[-1]

    assert start_lat == original_lat[start_index]
    assert end_lat == original_lat[end_index]

    # Longitudes
    original_lon = ds[lon_var].values
    lon = ds_filtered[lon_var].values
    start_lon = lon[0]
    end_lon = lon[-1]

    assert start_lon == original_lon[start_index]
    assert end_lon == original_lon[end_index]

    # Closest sample
    along_track_coords = get_coords(ds_filtered, lat_var=lat_var, lon_var=lon_var)
    distances = geodesic(along_track_coords, site_coords, units="km")
    closest_distance = np.min(distances)
    closest_filtered_index = np.argmin(np.abs(distances - closest_distance))
    closest_time = time[closest_filtered_index]
    closest_index = np.argmin(np.abs(original_time - closest_time))
    closest_lat = lat[closest_filtered_index]
    closest_lon = lat[closest_filtered_index]
    along_track_distance = get_cumulative_distances(lat, lon, units="km")[-1]

    assert start_time <= closest_time
    assert closest_time <= end_time
    assert start_index <= closest_index
    assert closest_index <= end_index

    # Pole crossing
    is_crossing_pole = ismonotonic(original_lat)

    # Site name
    if not isinstance(site_name, str):
        site_name = format_coords(lat=site_coords[0], lon=site_coords[1])

    return OverpassInfo(
        site_name=site_name,
        site_lat_deg_north=site_coords[0],
        site_lon_deg_east=site_coords[1],
        site_radius_km=site_radius_km,
        start_index=int(start_index),
        end_index=int(end_index),
        start_time=to_timestamp(start_time),
        end_time=to_timestamp(end_time),
        start_lat_deg_north=start_lat,
        start_lon_deg_east=start_lon,
        end_lat_deg_north=end_lat,
        end_lon_deg_east=end_lon,
        closest_index=int(closest_index),
        closest_lat_deg_north=closest_lat,
        closest_lon_deg_east=closest_lon,
        closest_time=to_timestamp(closest_time),
        closest_distance_km=float(closest_distance),
        along_track_distance_km=along_track_distance,
        frame_crosses_pole=is_crossing_pole,
        samples=len(time),
        site=_site,
    )


def get_overpass_info(
    product: str | xr.Dataset,
    site_radius_km: float | int,
    site: GroundSite | str,
    *,
    time_var: str = TIME_VAR,
    lat_var: str = TRACK_LAT_VAR,
    lon_var: str = TRACK_LON_VAR,
    along_track_dim: str = ALONG_TRACK_DIM,
) -> OverpassInfo:
    if isinstance(product, str):
        with read_product(product) as ds:
            result = _get_overpass_info(
                ds,
                site_radius_km=site_radius_km,
                site=site,
                time_var=time_var,
                lat_var=lat_var,
                lon_var=lon_var,
                along_track_dim=along_track_dim,
            )
    elif isinstance(product, xr.Dataset):
        result = _get_overpass_info(
            product,
            site_radius_km=site_radius_km,
            site=site,
            time_var=time_var,
            lat_var=lat_var,
            lon_var=lon_var,
            along_track_dim=along_track_dim,
        )
    else:
        raise TypeError(
            f"`product` has invalid type '{type(product).__name__}', expected 'str' (i.e. filepath) or 'xr.Dataset'"
        )

    return result



from dataclasses import dataclass
from enum import StrEnum


class FileInfoEnum(StrEnum):
    """Base class for file info enumerators."""

    @classmethod
    def from_input(cls, value: str):
        """Default implementation: Match input string to enum member (case insensitive)."""
        for member in cls:
            if member.value.lower() == value.lower():
                return member
            if member.name.lower() == value.lower():
                return member
        raise ValueError(
            f"Invalid value: {value}. Expected one of: {[m.value for m in cls]}"
        )

    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))



import logging
from typing import Iterable, Literal

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import xarray as xr
from matplotlib.axes import Axes
from matplotlib.colors import Colormap, LogNorm, Normalize
from matplotlib.dates import date2num
from matplotlib.figure import Figure, SubFigure
from numpy.typing import ArrayLike, NDArray

from ...utils.constants import *
from ...utils.profile_data import (
    ProfileData,
    ensure_along_track_2d,
    ensure_vertical_2d,
    validate_profile_data_dimensions,
)
from ...utils.swath_data import SwathData
from ...utils.swath_data.across_track_distance import get_nadir_index
from ...utils.time import TimeRangeLike, to_timestamp, validate_time_range
from ...utils.typing import DistanceRangeLike, ValueRangeLike
from ..color import Cmap, Color, get_cmap
from ..save import save_plot
from .along_track import AlongTrackAxisStyle, format_along_track_axis
from .colorbar import add_vertical_colorbar
from .defaults import get_default_cmap, get_default_norm, get_default_rolling_mean
from .height_ticks import format_height_ticks

logger = logging.getLogger(__name__)


class SwathFigure:
    def __init__(
        self,
        ax: Axes | None = None,
        figsize: tuple[float, float] = (12, 4),
        dpi: int | None = None,
        title: str | None = None,
    ):
        self.fig: Figure
        if isinstance(ax, Axes):
            tmp = ax.get_figure()
            if not isinstance(tmp, (Figure, SubFigure)):
                raise ValueError(f"Invalid Figure")
            self.fig = tmp  # type: ignore
            self.ax = ax
        else:
            # self.fig = plt.figure(figsize=figsize, dpi=dpi)
            # self.ax = self.fig.add_subplot()
            self.fig = plt.figure(figsize=figsize, dpi=dpi)
            self.ax = self.fig.add_axes((0.0, 0.0, 1.0, 1.0))

        self.title = title
        if self.title:
            self.fig.suptitle(self.title)

        self.ax_top: Axes | None = None
        self.ax_right: Axes | None = None
        self.selection_time_range: tuple[pd.Timestamp, pd.Timestamp] | None = None

    def plot(
        self,
        swath: SwathData | None = None,
        *,
        values: NDArray | None = None,
        time: NDArray | None = None,
        nadir_index: int | None = None,
        latitude: NDArray | None = None,
        longitude: NDArray | None = None,
        # Common args for wrappers
        value_range: ValueRangeLike | None = None,
        log_scale: bool | None = None,
        norm: Normalize | None = None,
        time_range: TimeRangeLike | None = None,
        from_track_range: DistanceRangeLike | None = None,
        label: str | None = None,
        units: str | None = None,
        cmap: str | Colormap | None = None,
        colorbar: bool = True,
        colorbar_ticks: ArrayLike | None = None,
        colorbar_tick_labels: ArrayLike | None = None,
        selection_time_range: TimeRangeLike | None = None,
        selection_color: str | None = Color("ec:earthcare"),
        selection_linestyle: str | None = "dashed",
        selection_linewidth: float | int | None = 2.5,
        selection_highlight: bool = False,
        selection_highlight_inverted: bool = True,
        selection_highlight_color: str = Color("white"),
        selection_highlight_alpha: float = 0.5,
        ax_style_top: AlongTrackAxisStyle | str = "geo",
        ax_style_bottom: AlongTrackAxisStyle | str = "time",
        ax_style_y: Literal[
            "from_track_distance", "across_track_distance", "pixel"
        ] = "from_track_distance",
        show_nadir: bool = True,
        **kwargs,
    ):
        if isinstance(value_range, Iterable):
            if len(value_range) != 2:
                raise ValueError(
                    f"invalid `value_range`: {value_range}, expecting (vmin, vmax)"
                )
        else:
            value_range = (None, None)

        cmap = get_cmap(cmap)

        if isinstance(cmap, Cmap) and cmap.categorical == True:
            norm = cmap.norm
        elif isinstance(norm, Normalize):
            if log_scale == True and not isinstance(norm, LogNorm):
                norm = LogNorm(norm.vmin, norm.vmax)
            elif log_scale == False and isinstance(norm, LogNorm):
                norm = Normalize(norm.vmin, norm.vmax)
            if value_range[0] is not None:
                norm.vmin = value_range[0]
            if value_range[1] is not None:
                norm.vmax = value_range[1]
        else:
            if log_scale == True:
                norm = LogNorm(value_range[0], value_range[1])
            else:
                norm = Normalize(value_range[0], value_range[1])

        assert isinstance(norm, Normalize)
        value_range = (norm.vmin, norm.vmax)

        if isinstance(swath, SwathData):
            values = swath.values
            time = swath.time
            nadir_index = swath.nadir_index
            latitude = swath.latitude
            longitude = swath.longitude
            label = swath.label
            units = swath.units
        elif (
            values is None
            or time is None
            or nadir_index is None
            or latitude is None
            or longitude is None
        ):
            raise ValueError(
                "Missing required arguments. Provide either a `SwathData` or all of `values`, `time`, `nadir_index`, `latitude` and `longitude`"
            )

        values = np.asarray(values)
        time = np.asarray(time)
        latitude = np.asarray(latitude)
        longitude = np.asarray(longitude)

        # Validate inputs
        logger.warning(f"Input validation not implemented")

        swath_data = SwathData(
            values=values,
            time=time,
            latitude=latitude,
            longitude=longitude,
            nadir_index=nadir_index,
            label=label,
            units=units,
        )

        tmin_original = swath_data.time[0]
        tmax_original = swath_data.time[-1]

        if from_track_range is not None:
            if isinstance(from_track_range, Iterable) and len(from_track_range) == 2:
                from_track_range = list(from_track_range)
                for i in [0, -1]:
                    if from_track_range[i] is None:
                        from_track_range[i] = swath_data.across_track_distance[i]
            swath_data = swath_data.select_from_track_range(from_track_range)
        else:
            from_track_range = (
                swath_data.across_track_distance[0],
                swath_data.across_track_distance[-1],
            )

        if time_range is not None:
            if isinstance(time_range, Iterable) and len(time_range) == 2:
                time_range = list(time_range)
                for i in [0, -1]:
                    if time_range[i] is None:
                        time_range[i] = to_timestamp(swath_data.time[i])
                    else:
                        time_range[i] = to_timestamp(time_range[i])
            swath_data = swath_data.select_time_range(time_range)
        else:
            time_range = (swath_data.time[0], swath_data.time[-1])

        values = swath_data.values
        time = swath_data.time
        latitude = swath_data.latitude
        longitude = swath_data.longitude
        across_track_distance = swath_data.across_track_distance
        from_track_distance = swath_data.from_track_distance
        label = swath_data.label
        units = swath_data.units
        nadir_index = swath_data.nadir_index

        is_from_track = ax_style_y != "across_track_distance"
        if ax_style_y == "from_track_distance":
            ydata = from_track_distance
            ylabel = "Distance from track"
        elif ax_style_y == "across_track_distance":
            ydata = across_track_distance
            ylabel = "Distance"
        elif ax_style_y == "pixel":
            ydata = np.arange(len(from_track_distance))
            ylabel = "Pixel"
        ynadir = ydata[nadir_index]

        tmin = np.datetime64(time_range[0])
        tmax = np.datetime64(time_range[1])

        if len(values.shape) == 3 and values.shape[2] == 3:
            mesh = self.ax.pcolormesh(time, ydata, values, **kwargs)
        else:
            mesh = self.ax.pcolormesh(
                time,
                ydata,
                values.T,
                norm=norm,
                cmap=cmap,
                **kwargs,
            )

            if colorbar:
                if cmap.categorical:
                    add_vertical_colorbar(
                        fig=self.fig,
                        ax=self.ax,
                        data=mesh,
                        label=f"{label} [{units}]",
                        ticks=cmap.ticks,
                        tick_labels=cmap.labels,
                    )
                else:
                    add_vertical_colorbar(
                        fig=self.fig,
                        ax=self.ax,
                        data=mesh,
                        label=f"{label} [{units}]",
                        ticks=colorbar_ticks,
                        tick_labels=colorbar_tick_labels,
                    )

        if selection_time_range is not None:
            self.selection_time_range = validate_time_range(selection_time_range)

            if selection_highlight:
                if selection_highlight_inverted:
                    self.ax.axvspan(  # type: ignore
                        tmin,  # type: ignore
                        self.selection_time_range[0],  # type: ignore
                        color=selection_highlight_color,  # type: ignore
                        alpha=selection_highlight_alpha,  # type: ignore
                    )  # type: ignore
                    self.ax.axvspan(  # type: ignore
                        self.selection_time_range[1],  # type: ignore
                        tmax,  # type: ignore
                        color=selection_highlight_color,  # type: ignore
                        alpha=selection_highlight_alpha,  # type: ignore
                    )  # type: ignore
                else:
                    self.ax.axvspan(  # type: ignore
                        self.selection_time_range[0],  # type: ignore
                        self.selection_time_range[1],  # type: ignore
                        color=selection_highlight_color,  # type: ignore
                        alpha=selection_highlight_alpha,  # type: ignore
                    )  # type: ignore

            for t in self.selection_time_range:
                self.ax.axvline(  # type: ignore
                    x=t,  # type: ignore
                    color=selection_color,  # type: ignore
                    linestyle=selection_linestyle,  # type: ignore
                    linewidth=selection_linewidth,  # type: ignore
                )  # type: ignore

        if show_nadir:
            self.ax.axhline(
                y=ynadir,
                color=Color("red"),
                linestyle="dashed",
                linewidth=2,
            )

        self.ax.set_xlim((tmin, tmax))  # type: ignore
        # self.ax.set_ylim((hmin, hmax))

        self.ax_right = self.ax.twinx()
        self.ax_right.set_ylim(self.ax.get_ylim())

        self.ax_top = self.ax.twiny()
        self.ax_top.set_xlim(self.ax.get_xlim())

        if ax_style_y == "pixel":
            format_height_ticks(self.ax, label=ylabel, show_units=False)
        else:
            format_height_ticks(self.ax, label=ylabel)
        format_height_ticks(
            self.ax_right, show_tick_labels=False, show_units=False, label=""
        )

        ax_style_bottom = AlongTrackAxisStyle.from_input(ax_style_bottom)
        ax_style_top = AlongTrackAxisStyle.from_input(ax_style_top)

        format_along_track_axis(
            self.ax,
            ax_style_bottom,
            time,
            tmin,
            tmax,
            tmin_original,
            tmax_original,
            longitude[:, nadir_index],
            latitude[:, nadir_index],
        )
        format_along_track_axis(
            self.ax_top,
            ax_style_top,
            time,
            tmin,
            tmax,
            tmin_original,
            tmax_original,
            longitude[:, nadir_index],
            latitude[:, nadir_index],
        )

    def ecplot(
        self,
        ds: xr.Dataset,
        var: str,
        *,
        time_var: str = TIME_VAR,
        lat_var: str = SWATH_LAT_VAR,
        lon_var: str = SWATH_LON_VAR,
        values: NDArray | None = None,
        time: NDArray | None = None,
        nadir_index: int | None = None,
        latitude: NDArray | None = None,
        longitude: NDArray | None = None,
        # Common args for wrappers
        value_range: ValueRangeLike | None = None,
        log_scale: bool | None = None,
        norm: Normalize | None = None,
        time_range: TimeRangeLike | None = None,
        from_track_range: DistanceRangeLike | None = None,
        label: str | None = None,
        units: str | None = None,
        cmap: str | Colormap | None = None,
        colorbar: bool = True,
        colorbar_ticks: ArrayLike | None = None,
        colorbar_tick_labels: ArrayLike | None = None,
        selection_time_range: TimeRangeLike | None = None,
        selection_color: str | None = Color("ec:earthcare"),
        selection_linestyle: str | None = "dashed",
        selection_linewidth: float | int | None = 2.5,
        selection_highlight: bool = False,
        selection_highlight_inverted: bool = True,
        selection_highlight_color: str = Color("white"),
        selection_highlight_alpha: float = 0.5,
        ax_style_top: AlongTrackAxisStyle | str = "geo",
        ax_style_bottom: AlongTrackAxisStyle | str = "time",
        ax_style_y: Literal[
            "from_track_distance", "across_track_distance", "pixel"
        ] = "from_track_distance",
        show_nadir: bool = True,
        **kwargs,
    ):
        # Collect all common args for wrapped plot function call
        local_args = locals()
        # Delete all args specific to this wrapper function
        del local_args["self"]
        del local_args["ds"]
        del local_args["var"]
        del local_args["time_var"]
        del local_args["lat_var"]
        del local_args["lon_var"]
        # Delete kwargs to then merge it with the residual common args
        del local_args["kwargs"]
        all_args = {**local_args, **kwargs}

        if all_args["values"] is None:
            all_args["values"] = ds[var].values
        if all_args["time"] is None:
            all_args["time"] = ds[time_var].values
        if all_args["nadir_index"] is None:
            all_args["nadir_index"] = get_nadir_index(ds)
        if all_args["latitude"] is None:
            all_args["latitude"] = ds[lat_var].values
        if all_args["longitude"] is None:
            all_args["longitude"] = ds[lon_var].values

        # Set default values depending on variable name
        if label is None:
            all_args["label"] = (
                "Values" if not hasattr(ds[var], "long_name") else ds[var].long_name
            )
        if units is None:
            all_args["units"] = "-" if not hasattr(ds[var], "units") else ds[var].units
        if value_range is None and log_scale is None and norm is None:
            all_args["norm"] = get_default_norm(var)
        if cmap is None:
            all_args["cmap"] = get_default_cmap(var)

        return self.plot(**all_args)

    def show(self):
        self.fig.tight_layout()
        self.fig.show()

    def save(self, filename: str = "", filepath: str | None = None, **kwargs):
        save_plot(fig=self.fig, filename=filename, filepath=filepath, **kwargs)



from .concat import concat_datasets
from .delete import remove_dims
from .exception import EmptyFilterResultError
from .filter_radius import filter_radius
from .filter_time import filter_time
from .merge import merge_datasets
from .scalars import convert_scalar_var_to_str



import logging

import xarray as xr


def _read_nc(
    filepath: str,
    modify: bool = True,
    **kwargs,
) -> xr.Dataset:
    """Returns an `xarray.Dataset` from a NetCDF file path.

    Args:
        filepath (str): Path to a NetCDF file.
        modify (bool): If True, default modifications to the opened dataset will be applied
            (e.g., converting heights in Polly data from height a.g.l. to height above mean sea level).
        **kwargs: Key-word arguments passed to `xarray.open_dataset()`.

    Returns:
        xarray.Dataset: The resulting dataset.

    Note:
        this function is basically a wrapper for the `xarray.open_dataset()` function.
    """

    logger = logging.getLogger()

    ds = xr.open_dataset(filepath, **kwargs)

    if modify:
        for var in ds.variables:
            units_attr: str | None = None
            if hasattr(ds[var], "units"):
                units_attr = "units"
            elif hasattr(ds[var], "unit"):
                units_attr = "unit"

            if isinstance(units_attr, str):
                units = ds[var].attrs[units_attr].lower()
                if "seconds" in units and "since" in units and "1970-01-01" in units:
                    ds[var].values = ds[var].values.astype("datetime64[s]")
                    ds[var].attrs[units_attr] = ""

        if (
            "altitude" in ds
            and "height" in ds
            and ds["altitude"].values.size == 1
            and ds["height"].values.size > 1
        ):
            logger.info(f"Convert height above ground level to height above ellipsoid.")
            ds["height"] = ds["height"] + ds["altitude"].values
            ds = ds.reset_index("height")
            # ds = ds.reset_coords("height")
            ds = ds.drop_vars("altitude")

    return ds


def read_nc(
    input: str | xr.Dataset,
    modify: bool = True,
    in_memory: bool = False,
    **kwargs,
) -> xr.Dataset:
    """Returns an `xarray.Dataset` from a Dataset or NetCDF file path, optionally loaded into memory.

    Args:
        input (xarray.Dataset or str): Path to a NetCDF file. If a already opened `xarray.Dataset` object is passed, it is returned as is.
        modify (bool): If True, default modifications to the opened dataset will be applied
            (e.g., converting heights in Polly data from height a.g.l. to height above mean sea level).
        in_memory (bool, optional): If True, ensures the dataset is fully loaded into memory. Defaults to False.
        **kwargs: Key-word arguments passed to `xarray.open_dataset()`.

    Returns:
        xarray.Dataset: The resulting dataset.

    Raises:
        TypeError: If input is not a Dataset or string.
    """
    ds: xr.Dataset
    if isinstance(input, xr.Dataset):
        ds = input
    elif isinstance(input, str):
        if in_memory:
            with _read_nc(input, modify=modify, **kwargs) as ds:
                ds = ds.load()
        else:
            ds = _read_nc(input, modify=modify, **kwargs)
    else:
        raise TypeError(
            f"Invalid input type! Expecting a opened NetCDF dataset (xarray.Dataset) or a path to a NetCDF file."
        )
    return ds



import numpy as np
import xarray as xr

from ..constants import (
    ACROSS_TRACK_DIM,
    ACROSS_TRACK_DISTANCE,
    FROM_TRACK_DISTANCE,
    SWATH_LAT_VAR,
    SWATH_LON_VAR,
    TRACK_LAT_VAR,
    TRACK_LON_VAR,
)
from ..geo import geodesic


def add_across_track_distance(
    ds: xr.Dataset,
    nadir_idx: int,
    swath_lat_var: str,
    swath_lon_var: str,
    across_track_distance_var: str = ACROSS_TRACK_DISTANCE,
    from_track_distance_var: str = FROM_TRACK_DISTANCE,
    across_track_dim: str = "across_track",
) -> xr.Dataset:
    """Extends EarthCARE dataset containing an across-track dimension by variable containing distance from nadir."""
    # Add across-track distance variable
    last_coords = np.vstack(
        (ds[swath_lat_var].values[:, -1], ds[swath_lon_var].values[:, -1])
    ).T
    nadir_coords = np.vstack(
        (ds[swath_lat_var].values[:, nadir_idx], ds[swath_lon_var].values[:, nadir_idx])
    ).T
    across_track_distances = np.array([], dtype=np.float32)
    from_track_distances = np.array([], dtype=np.float32)
    for i in range(0, ds[swath_lat_var].values.shape[1]):
        sign = 1 if i < nadir_idx else -1
        across_track_coords = np.vstack(
            (ds[swath_lat_var].values[:, i], ds[swath_lon_var].values[:, i])
        ).T

        _dists = geodesic(last_coords, across_track_coords, units="m")
        _mean_dists = np.mean(np.atleast_1d(_dists))
        across_track_distances = np.append(across_track_distances, _mean_dists)

        _dists = geodesic(nadir_coords, across_track_coords, units="m")
        _mean_dists = np.mean(np.atleast_1d(_dists)) * sign
        from_track_distances = np.append(from_track_distances, _mean_dists)

    ds[across_track_distance_var] = ((across_track_dim), across_track_distances)
    ds[across_track_distance_var] = ds[across_track_distance_var].assign_attrs(
        units="m", name="Distance", long_name="Distance"
    )

    ds[from_track_distance_var] = ((across_track_dim), from_track_distances)
    ds[from_track_distance_var] = ds[from_track_distance_var].assign_attrs(
        units="m", name="Distance from track", long_name="Distance from track"
    )
    # indices = np.arange(len(distances))
    # distances = np.interp(
    #     indices, indices[~np.isnan(distances)], distances[~np.isnan(distances)]
    # )
    # ds[across_track_distance_var].values = distances

    return ds


def add_nadir_track(
    ds: xr.Dataset,
    nadir_idx: int,
    swath_lat_var: str,
    swath_lon_var: str,
    along_track_dim: str,
    across_track_dim: str,
    nadir_lat_var: str = TRACK_LAT_VAR,
    nadir_lon_var: str = TRACK_LON_VAR,
) -> xr.Dataset:
    """Extends EarthCARE dataset containing an across-track dimension by nadir selected lat/lon variables."""

    if swath_lat_var == nadir_lat_var:
        raise ValueError(
            f"Track latitude and swath latitude variables must be different (lat_var={swath_lat_var}, nadir_lat_var={nadir_lat_var})"
        )

    if swath_lon_var == nadir_lon_var:
        raise ValueError(
            f"Track longitude and swath longitude variables must be different (lon_var={swath_lon_var}, nadir_lon_var={nadir_lon_var})"
        )

    # Drop all samples with missing geo location
    valid_across_track = ds[swath_lat_var].notnull().all(dim=along_track_dim)
    ds = ds.isel({across_track_dim: valid_across_track.values})

    # Add nadir track as lat/lon variables
    across_track_nadir_selection = {across_track_dim: nadir_idx}
    ds = ds.assign(
        {
            nadir_lat_var: ds[swath_lat_var].isel(across_track_nadir_selection),
            nadir_lon_var: ds[swath_lon_var].isel(across_track_nadir_selection),
        }
    )
    ds[nadir_lat_var] = ds[nadir_lat_var].assign_attrs(
        units="degree_north", notes="[-90:90]", long_name="Latitude"
    )
    ds[nadir_lon_var] = ds[nadir_lon_var].assign_attrs(
        units="degree_east", notes="[-180:180]", long_name="Longitude"
    )

    return ds


def add_nadir_var(
    ds: xr.Dataset,
    var: str,
    nadir_idx: int | None = None,
    new_var: str | None = None,
    across_track_dim: str = ACROSS_TRACK_DIM,
    units: str | None = None,
    notes: str | None = None,
    long_name: str | None = None,
) -> xr.Dataset:
    if not isinstance(new_var, str):
        new_var = f"{var}_track"

    if not isinstance(nadir_idx, int):
        nadir_idx = get_nadir_index(ds)

    if not isinstance(units, str) and hasattr(ds[var], "units"):
        units = ds[var].units

    if not isinstance(notes, str) and hasattr(ds[var], "notes"):
        notes = ds[var].notes

    if not isinstance(long_name, str) and hasattr(ds[var], "long_name"):
        long_name = ds[var].long_name

    # Add nadir track as lat/lon variables
    across_track_nadir_selection = {across_track_dim: nadir_idx}
    ds = ds.assign(
        {
            new_var: ds[var].isel(across_track_nadir_selection),
        }
    )
    ds[new_var] = ds[new_var].assign_attrs(
        units=units, notes=notes, long_name=long_name
    )

    return ds


def get_nadir_index(
    ds: xr.Dataset,
    nadir_idx: int | None = None,
    sensor_elevation_angle_var: str = "sensor_elevation_angle",
    across_track_dim: str = "across_track",
) -> int:
    """
    Gets the Nadir index in the across-track dimension debending on optional parameters given.

    Parameters:
        ds (xarray.Dataset):
            A EarthCARE dataset containing a along-track dimension.
        nadir_idx (int | None, optional):
            If given, the same index is returned. Defaults to None.
        sensor_elevation_angle_var (str, optional):
            The name in the dataset's sensor elevation angle variable. Defaults to 'sensor_elevation_angle'. Defaults to 'sensor_elevation_angle'.
        across_track_dim (str, optional):
            The name of the dataset's across-track dimension. Defaults to 'across_track'.

    Returns:
        nadir_index (int):
            The across-track index position of Nadir.
    """
    if nadir_idx is not None:
        return nadir_idx
    elif sensor_elevation_angle_var in ds.variables:
        return int(
            np.median(np.nanargmax(ds[sensor_elevation_angle_var].values, axis=1))
        )
    elif SWATH_LON_VAR in ds.variables and TRACK_LON_VAR in ds.variables:
        return int(
            np.nanargmin(
                np.abs(ds[SWATH_LON_VAR].values[0] - ds[TRACK_LON_VAR].values[0])
            )
        )
    return int(np.nanargmin(np.abs(ds[across_track_dim].values)))



import numpy as np
import xarray as xr
from matplotlib.colors import Colormap, LogNorm, Normalize

from ...utils import FileType
from ..color.colormap import Cmap, get_cmap


def get_default_norm(var: str, ds: xr.Dataset | None = None) -> Normalize:
    file_type: FileType | None = None
    if ds is not None and not isinstance(ds, FileType):
        file_type = FileType.from_input(ds)

    if var in [
        "mie_attenuated_backscatter",
        "rayleigh_attenuated_backscatter",
        "crosspolar_attenuated_backscatter",
    ]:
        return LogNorm(vmin=1e-8, vmax=1e-5)
    elif var in [
        "particle_backscatter_coefficient_355nm",
        "particle_backscatter_coefficient_355nm_medium_resolution",
        "particle_backscatter_coefficient_355nm_low_resolution",
    ]:
        return LogNorm(vmin=1e-7, vmax=1e-4)
    elif var in [
        "particle_extinction_coefficient_355nm",
        "particle_extinction_coefficient_355nm_medium_resolution",
        "particle_extinction_coefficient_355nm_low_resolution",
    ]:
        return LogNorm(vmin=1e-6, vmax=1e-3)
    elif var in [
        "lidar_ratio_355nm",
        "lidar_ratio_355nm_medium_resolution",
        "lidar_ratio_355nm_low_resolution",
    ]:
        return Normalize(vmin=0, vmax=100)
    elif var in [
        "particle_linear_depol_ratio_355nm",
        "particle_linear_depol_ratio_355nm_medium_resolution",
        "particle_linear_depol_ratio_355nm_low_resolution",
        "depol_ratio",
    ]:
        return Normalize(vmin=0, vmax=0.6)
    elif var in [
        "plot_radarReflectivityFactor",
    ]:
        return Normalize(vmin=-25, vmax=20)
    elif var in [
        "plot_dopplerVelocity",
    ]:
        return Normalize(vmin=-2, vmax=4)
    elif "cloud_top_height_MSI" in var:
        return Normalize(vmin=0)
    elif var == "plot_cloud_top_height_difference_ATLID_MSI" and isinstance(
        ds, xr.Dataset
    ):
        return Normalize(
            vmin=0, vmax=np.nanmax(ds["cloud_top_height_difference_ATLID_MSI"].values)
        )
    elif "cloud_top_height_difference_ATLID_MSI" in var:
        return Normalize(vmin=0)
    elif "quality_status" in var:
        return Normalize(vmin=-1.5, vmax=4.5)
    return Normalize()


def get_default_rolling_mean(
    var: str, file_type: str | xr.Dataset | FileType | None = None
) -> int | None:

    if file_type is not None and not isinstance(file_type, FileType):
        file_type = FileType.from_input(file_type)

    if var in [
        "mie_attenuated_backscatter",
        "rayleigh_attenuated_backscatter",
        "crosspolar_attenuated_backscatter",
    ]:
        return 20
    return None


def get_default_cmap(
    var: str, file_type: str | xr.Dataset | FileType | None = None
) -> Cmap:

    if file_type is not None and not isinstance(file_type, FileType):
        file_type = FileType.from_input(file_type)

    if var in [
        "mie_attenuated_backscatter",
        "crosspolar_attenuated_backscatter",
        "particle_backscatter_coefficient_355nm",
        "particle_backscatter_coefficient_355nm_medium_resolution",
        "particle_backscatter_coefficient_355nm_low_resolution",
    ]:
        return get_cmap("calipso")
    elif var in [
        "particle_extinction_coefficient_355nm",
        "particle_extinction_coefficient_355nm_medium_resolution",
        "particle_extinction_coefficient_355nm_low_resolution",
    ]:
        return get_cmap("chiljet2")
    elif var in [
        "lidar_ratio_355nm",
        "lidar_ratio_355nm_medium_resolution",
        "lidar_ratio_355nm_low_resolution",
    ]:
        return get_cmap("chiljet2")
    elif var in [
        "rayleigh_attenuated_backscatter",
    ]:
        return get_cmap("ray")
    elif var in [
        "particle_linear_depol_ratio_355nm",
        "particle_linear_depol_ratio_355nm_medium_resolution",
        "particle_linear_depol_ratio_355nm_low_resolution",
        "depol_ratio",
    ]:
        return get_cmap("ratio")
    elif var in [
        "simple_classification",
    ]:
        return get_cmap("simple_classification")
    elif var in [
        "classification",
        "classification_medium_resolution",
        "classification_low_resolution",
    ]:
        return get_cmap("target_classification")
    elif var in [
        "plot_radarReflectivityFactor",
    ]:
        return get_cmap("radar_reflectivity")
    elif var in [
        "plot_dopplerVelocity",
    ]:
        return get_cmap("doppler_velocity")
    elif "cloud_top_height_MSI" in var:
        return get_cmap(get_cmap("navia").with_extremes(bad="#ffffff00"))
    elif "cloud_top_height_difference_ATLID_MSI" in var:
        return get_cmap(get_cmap("navia").with_extremes(bad="#808080", over="white"))
    elif "quality_status" in var:
        if isinstance(file_type, FileType):
            if file_type == FileType.AM__CTH_2B:
                cmap = get_cmap("roma_r")
                colors = cmap(np.linspace(0.1, 1, 5))
                colors = np.append(np.array([[1, 1, 1, 1]]), colors, axis=0)
                definitions = {v: str(v) for v in [-1, 0, 1, 2, 3, 4]}
                cmap = Cmap(colors, name="quality_status_amcth").to_categorical(
                    definitions
                )
                return cmap
        cmap = get_cmap("roma_r")
        colors = cmap(np.linspace(0.1, 1, 5))
        colors = np.append(np.array([[1, 1, 1, 1]]), colors, axis=0)
        definitions = {v: str(v) for v in [-1, 0, 1, 2, 3, 4]}
        cmap = Cmap(colors, name="quality_status_amcth").to_categorical(definitions)
        return cmap
    return get_cmap("viridis")


def get_default_profile_range(
    var: str, ds: xr.Dataset | None = None
) -> tuple[float | None, float | None] | None:
    file_type: FileType | None = None
    if ds is not None and not isinstance(ds, FileType):
        file_type = FileType.from_input(ds)

    pad_frac = 0.00
    max_bsc = 8e-6  # [m-1 sr-1]
    max_ext = 6e-4  # [m-1]
    max_lr = 100.0  # [sr]
    max_depol = 0.5  # [-]

    _min: float
    _max: float
    if var in ["mie_attenuated_backscatter"]:
        _max = 0.8e-6
        return (-_max * pad_frac, _max)
    elif var in ["rayleigh_attenuated_backscatter"]:
        _max = 3e-6
        return (-_max * pad_frac, _max)
    elif var in ["crosspolar_attenuated_backscatter"]:
        _max = 0.2e-6
        return (-_max * pad_frac, _max)
    elif var in [
        "particle_backscatter_coefficient_355nm",
        "particle_backscatter_coefficient_355nm_medium_resolution",
        "particle_backscatter_coefficient_355nm_low_resolution",
    ]:
        _max = max_bsc
        return (-_max * pad_frac, _max)
    elif var in [
        "particle_extinction_coefficient_355nm",
        "particle_extinction_coefficient_355nm_medium_resolution",
        "particle_extinction_coefficient_355nm_low_resolution",
    ]:
        _max = max_ext
        return (-_max * pad_frac, _max)
    elif var in [
        "lidar_ratio_355nm",
        "lidar_ratio_355nm_medium_resolution",
        "lidar_ratio_355nm_low_resolution",
    ]:
        _max = max_lr
        return (-_max * pad_frac, _max)
    elif var in [
        "particle_linear_depol_ratio_355nm",
        "particle_linear_depol_ratio_355nm_medium_resolution",
        "particle_linear_depol_ratio_355nm_low_resolution",
        "depol_ratio",
    ]:
        _max = max_depol
        return (-_max * pad_frac, _max)
    elif var in [
        "plot_radarReflectivityFactor",
    ]:
        _min = -25
        _max = 20
        vrange = _max - _min
        pad = vrange * pad_frac
        return (_min - pad, _max + pad)
    elif var in [
        "plot_dopplerVelocity",
    ]:
        _min = -2
        _max = 4
        vrange = _max - _min
        pad = vrange * pad_frac
        return (_min - pad, _max + pad)
    return (None, None)



from pyproj import Geod

_geod = Geod(ellps="WGS84")


def interpgeo(
    lat1: float, lon1: float, lat2: float, lon2: float, f: float
) -> tuple[float, float]:
    """
    Interpolates along the geodesic from (lon1, lat1) to (lon2, lat2) by fraction f (0 to 1).
    Returns interpolated (lon, lat).
    """
    azi1, azi2, dist = _geod.inv(lon1, lat1, lon2, lat2)
    lon, lat, _ = _geod.fwd(lon1, lat1, azi1, f * dist)
    return lon, lat



import re


def parse_units(units: str) -> str:

    pattern = re.compile(
        r"(?i)^((mm|msr)(\^?-1)\s*(m|sr)(\^?-1)|(m|sr)(\^?-1)\s*(mm|msr)(\^?-1))$"
    )
    match = pattern.match(units)
    if match:
        return "Mm-1 sr-1"

    pattern = re.compile(r"(?i)^((m)(\^?-1)\s*(sr)(\^?-1)|(sr)(\^?-1)\s*(m)(\^?-1))$")
    match = pattern.match(units)
    if match:
        return "m-1 sr-1"

    pattern = re.compile(r"(?i)^mm(\^?-1)$")
    match = pattern.match(units)
    if match:
        return "Mm-1"

    pattern = re.compile(r"(?i)^m(\^?-1)$")
    match = pattern.match(units)
    if match:
        return "m-1"

    pattern = re.compile(r"(?i)^msr(\^?-1)$")
    match = pattern.match(units)
    if match:
        return "Msr-1"

    pattern = re.compile(r"(?i)^sr(\^?-1)$")
    match = pattern.match(units)
    if match:
        return "sr-1"

    return units.strip()



import numpy as np
from matplotlib.colors import Colormap, ListedColormap

from ..color import Color
from ..format_conversion import alpha_to_hex


def get_color_list_ectools(low_value_alpha: float = 1.0) -> list[tuple[int, Color]]:
    """
    Based on color index values from the calipso colormap in the 'ectools' project by Shannon Mason (ECMWF):
    https://bitbucket.org/smason/workspace/projects/EC
    Original under Apache 2.0 license.
    """
    return [
        (34, Color("#123598").set_alpha(low_value_alpha)),  # original value: #002BBB
        (76, Color("#007fff").set_alpha(low_value_alpha)),
        (79, Color("#00ffaa")),
        (82, Color("#007f7f")),
        (86, Color("#00aa55")),
        (101, Color("#ffff00")),
        (110, Color("#ffe600")),
        (119, Color("#ffd400")),
        (128, Color("#ffaa00")),
        (134, Color("#ff7f00")),
        (140, Color("#ff5500")),
        (146, Color("#ff0000")),
        (152, Color("#ff2a55")),
        (155, Color("#ff557f")),
        (158, Color("#ff7faa")),
        (161, Color("#464646")),
        (164, Color("#5a5a5a")),
        (167, Color("#6e6e6e")),
        (170, Color("#828282")),
        (176, Color("#969696")),
        (182, Color("#aaaaaa")),
        (188, Color("#b4b4b4")),
        (194, Color("#bebebe")),
        (200, Color("#c8c8c8")),
        (206, Color("#d2d2d2")),
        (212, Color("#d7d7d7")),
        (218, Color("#dcdcdc")),
        (224, Color("#e1e1e1")),
        (230, Color("#e6e6e6")),
        (236, Color("#ebebeb")),
        (243, Color("#f0f0f0")),
        (250, Color("#f5f5f5")),
        (257, Color("#ffffff")),
    ]


def get_color_list_old(low_value_alpha: float = 1.0) -> list[tuple[int, Color]]:
    return [
        (1, Color("#002BBB").set_alpha(low_value_alpha)),
        (2, Color("#002BBB").set_alpha(low_value_alpha)),
        (3, Color("#0B7DF7")),
        (4, Color("#06A7FF")),
        (5, Color("#02D0FF")),
        (6, Color("#13F5FE")),
        (7, Color("#09F8CF")),
        (8, Color("#02FCA8")),
        (9, Color("#107682")),
        (10, Color("#00A850")),
        (11, Color("#FAFF0F")),
        (12, Color("#FAFE04")),
        (13, Color("#FCD40B")),
        (14, Color("#FBAD03")),
        (15, Color("#F68200")),
        (16, Color("#EA5B00")),
        (17, Color("#E80800")),
        (18, Color("#FE2C53")),
        (19, Color("#F45C7F")),
        (20, Color("#FE7BA8")),
        (21, Color("#474747")),
        (22, Color("#656565")),
        (23, Color("#818181")),
        (24, Color("#9E9E9E")),
        (25, Color("#B6B6B6")),
        (26, Color("#C5C5C5")),
        (27, Color("#D5D5D5")),
        (28, Color("#EAEAEA")),
        (29, Color("#EEEEEE")),
        (30, Color("#F3F3F3")),
        (31, Color("#F6F6F6")),
        (32, Color("#F8F8F8")),
        (33, Color("#FCFCFC")),
        (34, Color("#FFFFFF")),
    ]


def get_cmap(low_value_alpha: float = 1.0) -> Colormap:
    """Creates the Calipso color map."""
    calipso_colors = get_color_list_ectools(low_value_alpha)

    bounds = np.append(0, [c[0] for c in calipso_colors])
    colors = np.array([c[1] for c in calipso_colors])
    bad_color = calipso_colors[0][1]

    color_list = np.concatenate(
        [[colors[i]] * (count) for i, count in enumerate(np.diff(bounds))]
    )
    cmap = ListedColormap(color_list, name="calipso").with_extremes(bad=bad_color)
    return cmap


def get_cmap_calipso_old(low_value_alpha: float = 1.0) -> Colormap:
    """Creates the Calipso color map."""
    calipso_colors = get_color_list_old(low_value_alpha)

    bounds = np.append(0, [c[0] for c in calipso_colors])
    colors = np.array([c[1] for c in calipso_colors])
    bad_color = calipso_colors[0][1]

    color_list = np.concatenate(
        [[colors[i]] * (count) for i, count in enumerate(np.diff(bounds))]
    )
    cmap = ListedColormap(color_list, name="calipso_old").with_extremes(bad=bad_color)
    return cmap
